# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_snapshot_metadata_proto/snapshot_metadata.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_tokenizer.proto import options_pb2 as pw__tokenizer_dot_proto_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_snapshot_metadata_proto/snapshot_metadata.proto',
  package='pw.snapshot',
  syntax='proto3',
  serialized_options=b'\n\021pw.snapshot.protoB\010Snapshot',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n2pw_snapshot_metadata_proto/snapshot_metadata.proto\x12\x0bpw.snapshot\x1a pw_tokenizer/proto/options.proto\"J\n\x0f\x43puArchitecture\"7\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x41RMV6M\x10\x01\x12\n\n\x06\x41RMV7M\x10\x02\x12\n\n\x06\x41RMV8M\x10\x03\"\xe9\x01\n\x08Metadata\x12\x14\n\x06reason\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\r\n\x05\x66\x61tal\x18\x02 \x01(\x08\x12\x1a\n\x0cproject_name\x18\x03 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x18\n\x10software_version\x18\x04 \x01(\t\x12\x1b\n\x13software_build_uuid\x18\x05 \x01(\x0c\x12\x19\n\x0b\x64\x65vice_name\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x15\n\rsnapshot_uuid\x18\x07 \x01(\x0c\x12\x33\n\x08\x63pu_arch\x18\x08 \x01(\x0e\x32!.pw.snapshot.CpuArchitecture.Enum\"\xa1\x01\n\x11SnapshotBasicInfo\x12\'\n\x08metadata\x18\x10 \x01(\x0b\x32\x15.pw.snapshot.Metadata\x12\x36\n\x04tags\x18\x11 \x03(\x0b\x32(.pw.snapshot.SnapshotBasicInfo.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x1d\n\x11pw.snapshot.protoB\x08Snapshotb\x06proto3'
  ,
  dependencies=[pw__tokenizer_dot_proto_dot_options__pb2.DESCRIPTOR,])



_CPUARCHITECTURE_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.snapshot.CpuArchitecture.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ARMV6M', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ARMV7M', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ARMV8M', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=120,
  serialized_end=175,
)
_sym_db.RegisterEnumDescriptor(_CPUARCHITECTURE_ENUM)


_CPUARCHITECTURE = _descriptor.Descriptor(
  name='CpuArchitecture',
  full_name='pw.snapshot.CpuArchitecture',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CPUARCHITECTURE_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=175,
)


_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='pw.snapshot.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='reason', full_name='pw.snapshot.Metadata.reason', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fatal', full_name='pw.snapshot.Metadata.fatal', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_name', full_name='pw.snapshot.Metadata.project_name', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='software_version', full_name='pw.snapshot.Metadata.software_version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='software_build_uuid', full_name='pw.snapshot.Metadata.software_build_uuid', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_name', full_name='pw.snapshot.Metadata.device_name', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_uuid', full_name='pw.snapshot.Metadata.snapshot_uuid', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpu_arch', full_name='pw.snapshot.Metadata.cpu_arch', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=178,
  serialized_end=411,
)


_SNAPSHOTBASICINFO_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='pw.snapshot.SnapshotBasicInfo.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.snapshot.SnapshotBasicInfo.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.snapshot.SnapshotBasicInfo.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=532,
  serialized_end=575,
)

_SNAPSHOTBASICINFO = _descriptor.Descriptor(
  name='SnapshotBasicInfo',
  full_name='pw.snapshot.SnapshotBasicInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='pw.snapshot.SnapshotBasicInfo.metadata', index=0,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='pw.snapshot.SnapshotBasicInfo.tags', index=1,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SNAPSHOTBASICINFO_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=414,
  serialized_end=575,
)

_CPUARCHITECTURE_ENUM.containing_type = _CPUARCHITECTURE
_METADATA.fields_by_name['cpu_arch'].enum_type = _CPUARCHITECTURE_ENUM
_SNAPSHOTBASICINFO_TAGSENTRY.containing_type = _SNAPSHOTBASICINFO
_SNAPSHOTBASICINFO.fields_by_name['metadata'].message_type = _METADATA
_SNAPSHOTBASICINFO.fields_by_name['tags'].message_type = _SNAPSHOTBASICINFO_TAGSENTRY
DESCRIPTOR.message_types_by_name['CpuArchitecture'] = _CPUARCHITECTURE
DESCRIPTOR.message_types_by_name['Metadata'] = _METADATA
DESCRIPTOR.message_types_by_name['SnapshotBasicInfo'] = _SNAPSHOTBASICINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CpuArchitecture = _reflection.GeneratedProtocolMessageType('CpuArchitecture', (_message.Message,), {
  'DESCRIPTOR' : _CPUARCHITECTURE,
  '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
  # @@protoc_insertion_point(class_scope:pw.snapshot.CpuArchitecture)
  })
_sym_db.RegisterMessage(CpuArchitecture)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
  'DESCRIPTOR' : _METADATA,
  '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
  # @@protoc_insertion_point(class_scope:pw.snapshot.Metadata)
  })
_sym_db.RegisterMessage(Metadata)

SnapshotBasicInfo = _reflection.GeneratedProtocolMessageType('SnapshotBasicInfo', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SNAPSHOTBASICINFO_TAGSENTRY,
    '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
    # @@protoc_insertion_point(class_scope:pw.snapshot.SnapshotBasicInfo.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _SNAPSHOTBASICINFO,
  '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
  # @@protoc_insertion_point(class_scope:pw.snapshot.SnapshotBasicInfo)
  })
_sym_db.RegisterMessage(SnapshotBasicInfo)
_sym_db.RegisterMessage(SnapshotBasicInfo.TagsEntry)


DESCRIPTOR._options = None
_METADATA.fields_by_name['reason']._options = None
_METADATA.fields_by_name['project_name']._options = None
_METADATA.fields_by_name['device_name']._options = None
_SNAPSHOTBASICINFO_TAGSENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""SearchToolbar class used by LogPanes."""

from __future__ import annotations
import functools
from typing import TYPE_CHECKING

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import Condition, has_focus
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.key_binding import (
    KeyBindings,
    KeyBindingsBase,
    KeyPressEvent,
)
from prompt_toolkit.layout import (
    ConditionalContainer,
    FormattedTextControl,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.widgets import TextArea
from prompt_toolkit.validation import DynamicValidator

from pw_console.log_view import RegexValidator, SearchMatcher
from pw_console.widgets import (
    mouse_handlers,
    to_checkbox_with_keybind_indicator,
    to_keybind_indicator,
)

if TYPE_CHECKING:
    from pw_console.log_pane import LogPane


class SearchToolbar(ConditionalContainer):
    """Toolbar for entering search text and viewing match counts."""

    TOOLBAR_HEIGHT = 2

    def __init__(self, log_pane: 'LogPane'):
        self.log_pane = log_pane
        self.log_view = log_pane.log_view
        self.search_validator = RegexValidator()
        self._search_successful = False
        self._search_invert = False
        self._search_field = None

        self.input_field = TextArea(
            prompt=[
                (
                    'class:search-bar-setting',
                    '/',
                    functools.partial(
                        mouse_handlers.on_click,
                        self.focus_self,
                    ),
                )
            ],
            focusable=True,
            focus_on_click=True,
            scrollbar=False,
            multiline=False,
            height=1,
            dont_extend_height=True,
            dont_extend_width=False,
            accept_handler=self._search_accept_handler,
            validator=DynamicValidator(self.get_search_matcher),
            history=self.log_pane.application.search_history,
        )

        self.input_field.control.key_bindings = self._create_key_bindings()

        match_count_window = Window(
            content=FormattedTextControl(self.get_match_count_fragments),
            height=1,
            align=WindowAlign.LEFT,
            dont_extend_width=True,
            style='class:search-match-count-dialog',
        )

        match_buttons_window = Window(
            content=FormattedTextControl(self.get_button_fragments),
            height=1,
            align=WindowAlign.LEFT,
            dont_extend_width=False,
            style='class:search-match-count-dialog',
        )

        input_field_buttons_window = Window(
            content=FormattedTextControl(self.get_search_help_fragments),
            height=1,
            align=WindowAlign.RIGHT,
            dont_extend_width=True,
        )

        settings_bar_window = Window(
            content=FormattedTextControl(self.get_search_settings_fragments),
            height=1,
            align=WindowAlign.LEFT,
            dont_extend_width=False,
        )

        super().__init__(
            HSplit(
                [
                    # Top row
                    VSplit(
                        [
                            # Search Settings toggles, only show if the search
                            # input field is in focus.
                            ConditionalContainer(
                                settings_bar_window,
                                filter=has_focus(self.input_field),
                            ),
                            # Match count numbers and buttons, only show if the
                            # search input is NOT in focus.
                            # pylint: disable=invalid-unary-operand-type
                            ConditionalContainer(
                                match_count_window,
                                filter=~has_focus(self.input_field),
                            ),
                            ConditionalContainer(
                                match_buttons_window,
                                filter=~has_focus(self.input_field),
                            ),
                            # pylint: enable=invalid-unary-operand-type
                        ]
                    ),
                    # Bottom row
                    VSplit(
                        [
                            self.input_field,
                            ConditionalContainer(
                                input_field_buttons_window,
                                filter=has_focus(self),
                            ),
                        ]
                    ),
                ],
                height=SearchToolbar.TOOLBAR_HEIGHT,
                style='class:search-bar',
            ),
            filter=Condition(lambda: log_pane.search_bar_active),
        )

    def _create_key_bindings(self) -> KeyBindingsBase:
        """Create additional key bindings for the search input."""
        # Clear filter keybind is handled by the parent log_pane.

        key_bindings = KeyBindings()
        register = self.log_pane.application.prefs.register_keybinding

        @register('search-toolbar.cancel', key_bindings)
        def _close_search_bar(_event: KeyPressEvent) -> None:
            """Close search bar."""
            self.cancel_search()

        @register('search-toolbar.toggle-matcher', key_bindings)
        def _select_next_search_matcher(_event: KeyPressEvent) -> None:
            """Select the next search matcher."""
            self.log_pane.log_view.select_next_search_matcher()

        @register('search-toolbar.create-filter', key_bindings)
        def _create_filter(_event: KeyPressEvent) -> None:
            """Create a filter."""
            self.create_filter()

        @register('search-toolbar.toggle-invert', key_bindings)
        def _toggle_search_invert(_event: KeyPressEvent) -> None:
            """Toggle inverted search matching."""
            self._invert_search()

        @register('search-toolbar.toggle-column', key_bindings)
        def _select_next_field(_event: KeyPressEvent) -> None:
            """Select next search field/column."""
            self._next_field()

        return key_bindings

    def focus_self(self) -> None:
        self.log_pane.application.application.layout.focus(self)

    def focus_log_pane(self) -> None:
        self.log_pane.application.focus_on_container(self.log_pane)

    def _create_filter(self) -> None:
        self.input_field.buffer.reset()
        self.close_search_bar()
        self.log_view.apply_filter()

    def _next_match(self) -> None:
        self.log_view.search_forwards()

    def _previous_match(self) -> None:
        self.log_view.search_backwards()

    def cancel_search(self) -> None:
        self.input_field.buffer.reset()
        self.close_search_bar()
        self.log_view.clear_search()

    def close_search_bar(self) -> None:
        """Close search bar."""
        # Reset invert setting for the next search
        self._search_invert = False
        self.log_view.follow_search_match = False
        # Hide the search bar
        self.log_pane.search_bar_active = False
        # Focus on the log_pane.
        self.log_pane.application.focus_on_container(self.log_pane)
        self.log_pane.redraw_ui()

    def _start_search(self) -> None:
        self.input_field.buffer.validate_and_handle()

    def _invert_search(self) -> None:
        self._search_invert = not self._search_invert

    def _toggle_search_follow(self) -> None:
        self.log_view.follow_search_match = (
            not self.log_view.follow_search_match
        )
        # If automatically jumping to the next search match, disable normal
        # follow mode.
        if self.log_view.follow_search_match:
            self.log_view.follow = False

    def _next_field(self) -> None:
        fields = self.log_pane.log_view.log_store.table.all_column_names()
        fields.append(None)
        current_index = fields.index(self._search_field)
        next_index = (current_index + 1) % len(fields)
        self._search_field = fields[next_index]

    def create_filter(self) -> None:
        self._start_search()
        if self._search_successful:
            self.log_pane.log_view.apply_filter()

    def _search_accept_handler(self, buff: Buffer) -> bool:
        """Function run when hitting Enter in the search bar."""
        self._search_successful = False
        if len(buff.text) == 0:
            self.close_search_bar()
            # Don't apply an empty search.
            return False

        if self.log_pane.log_view.new_search(
            buff.text, invert=self._search_invert, field=self._search_field
        ):
            self._search_successful = True

            # Don't close the search bar, instead focus on the log content.
            self.log_pane.application.focus_on_container(
                self.log_pane.log_display_window
            )
            # Keep existing search text.
            return True

        # Keep existing text if regex error
        return True

    def get_search_help_fragments(self):
        """Return FormattedText with search general help keybinds."""
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        start_search = functools.partial(
            mouse_handlers.on_click, self._start_search
        )
        close_search = functools.partial(
            mouse_handlers.on_click, self.cancel_search
        )

        # Search toolbar is darker than pane toolbars, use the darker button
        # style here.
        button_style = 'class:toolbar-button-inactive'

        separator_text = [('', '  ', focus)]

        # Empty text matching the width of the search bar title.
        fragments = [
            ('', '        ', focus),
        ]
        fragments.extend(separator_text)

        fragments.extend(
            to_keybind_indicator(
                'Enter', 'Search', start_search, base_style=button_style
            )
        )
        fragments.extend(separator_text)

        fragments.extend(
            to_keybind_indicator(
                'Ctrl-c', 'Cancel', close_search, base_style=button_style
            )
        )

        return fragments

    def get_search_settings_fragments(self):
        """Return FormattedText with current search settings and keybinds."""
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        next_field = functools.partial(
            mouse_handlers.on_click, self._next_field
        )
        toggle_invert = functools.partial(
            mouse_handlers.on_click, self._invert_search
        )
        next_matcher = functools.partial(
            mouse_handlers.on_click,
            self.log_pane.log_view.select_next_search_matcher,
        )

        separator_text = [('', '  ', focus)]

        # Search toolbar is darker than pane toolbars, use the darker button
        # style here.
        button_style = 'class:toolbar-button-inactive'

        fragments = [
            # Title
            ('class:search-bar-title', ' Search ', focus),
        ]
        fragments.extend(separator_text)

        selected_column_text = [
            (
                button_style + ' class:search-bar-setting',
                (self._search_field.title() if self._search_field else 'All'),
                next_field,
            ),
        ]
        fragments.extend(
            to_keybind_indicator(
                'Ctrl-t',
                'Column:',
                next_field,
                middle_fragments=selected_column_text,
                base_style=button_style,
            )
        )
        fragments.extend(separator_text)

        fragments.extend(
            to_checkbox_with_keybind_indicator(
                self._search_invert,
                'Ctrl-v',
                'Invert',
                toggle_invert,
                base_style=button_style,
            )
        )
        fragments.extend(separator_text)

        # Matching Method
        current_matcher_text = [
            (
                button_style + ' class:search-bar-setting',
                str(self.log_pane.log_view.search_matcher.name),
                next_matcher,
            )
        ]
        fragments.extend(
            to_keybind_indicator(
                'Ctrl-n',
                'Matcher:',
                next_matcher,
                middle_fragments=current_matcher_text,
                base_style=button_style,
            )
        )
        fragments.extend(separator_text)

        return fragments

    def get_search_matcher(self):
        if self.log_pane.log_view.search_matcher == SearchMatcher.REGEX:
            return self.log_pane.log_view.search_validator
        return False

    def get_match_count_fragments(self):
        """Return formatted text for the match count indicator."""
        focus = functools.partial(mouse_handlers.on_click, self.focus_log_pane)
        two_spaces = ('', '  ', focus)

        # Check if this line is a search match
        match_number = self.log_view.search_matched_lines.get(
            self.log_view.log_index, -1
        )

        # If valid, increment the zero indexed value by one for better human
        # readability.
        if match_number >= 0:
            match_number += 1
        # If no match, mark as zero
        else:
            match_number = 0

        return [
            ('class:search-match-count-dialog-title', ' Match ', focus),
            (
                '',
                '{} / {}'.format(
                    match_number, len(self.log_view.search_matched_lines)
                ),
                focus,
            ),
            two_spaces,
        ]

    def get_button_fragments(self) -> StyleAndTextTuples:
        """Return formatted text for the action buttons."""
        focus = functools.partial(mouse_handlers.on_click, self.focus_log_pane)

        one_space = ('', ' ', focus)
        two_spaces = ('', '  ', focus)
        cancel = functools.partial(mouse_handlers.on_click, self.cancel_search)
        create_filter = functools.partial(
            mouse_handlers.on_click, self._create_filter
        )
        next_match = functools.partial(
            mouse_handlers.on_click, self._next_match
        )
        previous_match = functools.partial(
            mouse_handlers.on_click, self._previous_match
        )
        toggle_search_follow = functools.partial(
            mouse_handlers.on_click,
            self._toggle_search_follow,
        )

        button_style = 'class:toolbar-button-inactive'

        fragments = []
        fragments.extend(
            to_keybind_indicator(
                key='n',
                description='Next',
                mouse_handler=next_match,
                base_style=button_style,
            )
        )
        fragments.append(two_spaces)

        fragments.extend(
            to_keybind_indicator(
                key='N',
                description='Previous',
                mouse_handler=previous_match,
                base_style=button_style,
            )
        )
        fragments.append(two_spaces)

        fragments.extend(
            to_keybind_indicator(
                key='Ctrl-c',
                description='Cancel',
                mouse_handler=cancel,
                base_style=button_style,
            )
        )
        fragments.append(two_spaces)

        fragments.extend(
            to_keybind_indicator(
                key='Ctrl-Alt-f',
                description='Add Filter',
                mouse_handler=create_filter,
                base_style=button_style,
            )
        )
        fragments.append(two_spaces)

        fragments.extend(
            to_checkbox_with_keybind_indicator(
                checked=self.log_view.follow_search_match,
                key='',
                description='Jump to new matches',
                mouse_handler=toggle_search_follow,
                base_style=button_style,
            )
        )
        fragments.append(one_space)

        return fragments



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_rpc protoc plugin entrypoint to generate code for RPC services."""

import enum
import sys

from google.protobuf.compiler import plugin_pb2

from pw_rpc import codegen_nanopb
from pw_rpc import codegen_pwpb
from pw_rpc import codegen_raw


class Codegen(enum.Enum):
    RAW = 0
    NANOPB = 1
    PWPB = 2


def process_proto_request(
    codegen: Codegen,
    req: plugin_pb2.CodeGeneratorRequest,
    res: plugin_pb2.CodeGeneratorResponse,
) -> None:
    """Handles a protoc CodeGeneratorRequest message.

    Generates code for the files in the request and writes the output to the
    specified CodeGeneratorResponse message.

    Args:
      req: A CodeGeneratorRequest for a proto compilation.
      res: A CodeGeneratorResponse to populate with the plugin's output.
    """
    for proto_file in req.proto_file:
        if codegen is Codegen.RAW:
            output_files = codegen_raw.process_proto_file(proto_file)
        elif codegen is Codegen.NANOPB:
            output_files = codegen_nanopb.process_proto_file(proto_file)
        elif codegen is Codegen.PWPB:
            output_files = codegen_pwpb.process_proto_file(proto_file)
        else:
            raise NotImplementedError(f'Unknown codegen type {codegen}')

        for output_file in output_files:
            fd = res.file.add()
            fd.name = output_file.name()
            fd.content = output_file.content()


def main(codegen: Codegen) -> int:
    """Protobuf compiler plugin entrypoint.

    Reads a CodeGeneratorRequest proto from stdin and writes a
    CodeGeneratorResponse to stdout.
    """
    data = sys.stdin.buffer.read()
    request = plugin_pb2.CodeGeneratorRequest.FromString(data)
    response = plugin_pb2.CodeGeneratorResponse()
    process_proto_request(codegen, request, response)

    # Declare that this plugin supports optional fields in proto3. No proto
    # message code is generated, so optional in proto3 is supported trivially.
    response.supported_features |= (  # type: ignore[attr-defined]
        response.FEATURE_PROTO3_OPTIONAL
    )  # type: ignore[attr-defined]

    sys.stdout.buffer.write(response.SerializeToString())
    return 0



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
r"""Decodes and detokenizes strings from binary or Base64 input.

The main class provided by this module is the Detokenize class. To use it,
construct it with the path to an ELF or CSV database, a tokens.Database,
or a file object for an ELF file or CSV. Then, call the detokenize method with
encoded messages, one at a time. The detokenize method returns a
DetokenizedString object with the result.

For example,

  from pw_tokenizer import detokenize

  detok = detokenize.Detokenizer('path/to/my/image.elf')
  print(detok.detokenize(b'\x12\x34\x56\x78\x03hi!'))

This module also provides a command line interface for decoding and detokenizing
messages from a file or stdin.
"""

import argparse
import base64
import binascii
import io
import logging
import os
from pathlib import Path
import re
import string
import struct
import sys
import time
from typing import (
    AnyStr,
    BinaryIO,
    Callable,
    Dict,
    List,
    Iterable,
    Iterator,
    Match,
    NamedTuple,
    Optional,
    Pattern,
    Tuple,
    Union,
)

try:
    from pw_tokenizer import database, decode, encode, tokens
except ImportError:
    # Append this path to the module search path to allow running this module
    # without installing the pw_tokenizer package.
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from pw_tokenizer import database, decode, encode, tokens

_LOG = logging.getLogger('pw_tokenizer')

ENCODED_TOKEN = struct.Struct('<I')
BASE64_PREFIX = encode.BASE64_PREFIX.encode()
DEFAULT_RECURSION = 9

_RawIO = Union[io.RawIOBase, BinaryIO]


class DetokenizedString:
    """A detokenized string, with all results if there are collisions."""

    def __init__(
        self,
        token: Optional[int],
        format_string_entries: Iterable[tuple],
        encoded_message: bytes,
        show_errors: bool = False,
    ):
        self.token = token
        self.encoded_message = encoded_message
        self._show_errors = show_errors

        self.successes: List[decode.FormattedString] = []
        self.failures: List[decode.FormattedString] = []

        decode_attempts: List[Tuple[Tuple, decode.FormattedString]] = []

        for entry, fmt in format_string_entries:
            result = fmt.format(
                encoded_message[ENCODED_TOKEN.size :], show_errors
            )
            decode_attempts.append((result.score(entry.date_removed), result))

        # Sort the attempts by the score so the most likely results are first.
        decode_attempts.sort(key=lambda value: value[0], reverse=True)

        # Split out the successesful decodes from the failures.
        for score, result in decode_attempts:
            if score[0]:
                self.successes.append(result)
            else:
                self.failures.append(result)

    def ok(self) -> bool:
        """True if exactly one string decoded the arguments successfully."""
        return len(self.successes) == 1

    def matches(self) -> List[decode.FormattedString]:
        """Returns the strings that matched the token, best matches first."""
        return self.successes + self.failures

    def best_result(self) -> Optional[decode.FormattedString]:
        """Returns the string and args for the most likely decoded string."""
        for string_and_args in self.matches():
            return string_and_args

        return None

    def error_message(self) -> str:
        """If detokenization failed, returns a descriptive message."""
        if self.ok():
            return ''

        if not self.matches():
            if self.token is None:
                return 'missing token'

            return 'unknown token {:08x}'.format(self.token)

        if len(self.matches()) == 1:
            return 'decoding failed for {!r}'.format(self.matches()[0].value)

        return '{} matches'.format(len(self.matches()))

    def __str__(self) -> str:
        """Returns the string for the most likely result."""
        result = self.best_result()
        if result:
            return result[0]

        if self._show_errors:
            return '<[ERROR: {}|{!r}]>'.format(
                self.error_message(), self.encoded_message
            )

        # Display the string as prefixed Base64 if it cannot be decoded.
        return encode.prefixed_base64(self.encoded_message)

    def __repr__(self) -> str:
        if self.ok():
            message = repr(str(self))
        else:
            message = 'ERROR: {}|{!r}'.format(
                self.error_message(), self.encoded_message
            )

        return '{}({})'.format(type(self).__name__, message)


class _TokenizedFormatString(NamedTuple):
    entry: tokens.TokenizedStringEntry
    format: decode.FormatString


class Detokenizer:
    """Main detokenization class; detokenizes strings and caches results."""

    def __init__(self, *token_database_or_elf, show_errors: bool = False):
        """Decodes and detokenizes binary messages.

        Args:
          *token_database_or_elf: a path or file object for an ELF or CSV
              database, a tokens.Database, or an elf_reader.Elf
          show_errors: if True, an error message is used in place of the %
              conversion specifier when an argument fails to decode
        """
        self.show_errors = show_errors

        # Cache FormatStrings for faster lookup & formatting.
        self._cache: Dict[int, List[_TokenizedFormatString]] = {}

        self._initialize_database(token_database_or_elf)

    def _initialize_database(self, token_sources: Iterable) -> None:
        self.database = database.load_token_database(*token_sources)
        self._cache.clear()

    def lookup(self, token: int) -> List[_TokenizedFormatString]:
        """Returns (TokenizedStringEntry, FormatString) list for matches."""
        try:
            return self._cache[token]
        except KeyError:
            format_strings = [
                _TokenizedFormatString(entry, decode.FormatString(str(entry)))
                for entry in self.database.token_to_entries[token]
            ]
            self._cache[token] = format_strings
            return format_strings

    def detokenize(self, encoded_message: bytes) -> DetokenizedString:
        """Decodes and detokenizes a message as a DetokenizedString."""
        if not encoded_message:
            return DetokenizedString(
                None, (), encoded_message, self.show_errors
            )

        # Pad messages smaller than ENCODED_TOKEN.size with zeroes to support
        # tokens smaller than a uint32. Messages with arguments must always use
        # a full 32-bit token.
        missing_token_bytes = ENCODED_TOKEN.size - len(encoded_message)
        if missing_token_bytes > 0:
            encoded_message += b'\0' * missing_token_bytes

        (token,) = ENCODED_TOKEN.unpack_from(encoded_message)
        return DetokenizedString(
            token, self.lookup(token), encoded_message, self.show_errors
        )

    def detokenize_base64(
        self,
        data: AnyStr,
        prefix: Union[str, bytes] = BASE64_PREFIX,
        recursion: int = DEFAULT_RECURSION,
    ) -> AnyStr:
        """Decodes and replaces prefixed Base64 messages in the provided data.

        Args:
          data: the binary data to decode
          prefix: one-character byte string that signals the start of a message
          recursion: how many levels to recursively decode

        Returns:
          copy of the data with all recognized tokens decoded
        """
        output = io.BytesIO()
        self.detokenize_base64_to_file(data, output, prefix, recursion)
        result = output.getvalue()
        return result.decode() if isinstance(data, str) else result

    def detokenize_base64_to_file(
        self,
        data: Union[str, bytes],
        output: BinaryIO,
        prefix: Union[str, bytes] = BASE64_PREFIX,
        recursion: int = DEFAULT_RECURSION,
    ) -> None:
        """Decodes prefixed Base64 messages in data; decodes to output file."""
        data = data.encode() if isinstance(data, str) else data
        prefix = prefix.encode() if isinstance(prefix, str) else prefix

        output.write(
            _base64_message_regex(prefix).sub(
                self._detokenize_prefixed_base64(prefix, recursion), data
            )
        )

    def detokenize_base64_live(
        self,
        input_file: _RawIO,
        output: BinaryIO,
        prefix: Union[str, bytes] = BASE64_PREFIX,
        recursion: int = DEFAULT_RECURSION,
    ) -> None:
        """Reads chars one-at-a-time, decoding messages; SLOW for big files."""
        prefix_bytes = prefix.encode() if isinstance(prefix, str) else prefix

        base64_message = _base64_message_regex(prefix_bytes)

        def transform(data: bytes) -> bytes:
            return base64_message.sub(
                self._detokenize_prefixed_base64(prefix_bytes, recursion), data
            )

        for message in PrefixedMessageDecoder(
            prefix, string.ascii_letters + string.digits + '+/-_='
        ).transform(input_file, transform):
            output.write(message)

            # Flush each line to prevent delays when piping between processes.
            if b'\n' in message:
                output.flush()

    def _detokenize_prefixed_base64(
        self, prefix: bytes, recursion: int
    ) -> Callable[[Match[bytes]], bytes]:
        """Returns a function that decodes prefixed Base64."""

        def decode_and_detokenize(match: Match[bytes]) -> bytes:
            """Decodes prefixed base64 with this detokenizer."""
            original = match.group(0)

            try:
                detokenized_string = self.detokenize(
                    base64.b64decode(original[1:], validate=True)
                )
                if detokenized_string.matches():
                    result = str(detokenized_string).encode()

                    if recursion > 0 and original != result:
                        result = self.detokenize_base64(
                            result, prefix, recursion - 1
                        )

                    return result
            except binascii.Error:
                pass

            return original

        return decode_and_detokenize


_PathOrStr = Union[Path, str]


# TODO(b/265334753): Reuse this function in database.py:LoadTokenDatabases
def _parse_domain(path: _PathOrStr) -> Tuple[Path, Optional[Pattern[str]]]:
    """Extracts an optional domain regex pattern suffix from a path"""

    if isinstance(path, Path):
        path = str(path)

    delimiters = path.count('#')

    if delimiters == 0:
        return Path(path), None

    if delimiters == 1:
        path, domain = path.split('#')
        return Path(path), re.compile(domain)

    raise ValueError(
        f'Too many # delimiters. Expected 0 or 1, found {delimiters}'
    )


class AutoUpdatingDetokenizer(Detokenizer):
    """Loads and updates a detokenizer from database paths."""

    class _DatabasePath:
        """Tracks the modified time of a path or file object."""

        def __init__(self, path: _PathOrStr) -> None:
            self.path, self.domain = _parse_domain(path)
            self._modified_time: Optional[float] = self._last_modified_time()

        def updated(self) -> bool:
            """True if the path has been updated since the last call."""
            modified_time = self._last_modified_time()
            if modified_time is None or modified_time == self._modified_time:
                return False

            self._modified_time = modified_time
            return True

        def _last_modified_time(self) -> Optional[float]:
            try:
                return os.path.getmtime(self.path)
            except FileNotFoundError:
                return None

        def load(self) -> tokens.Database:
            try:
                if self.domain is not None:
                    return database.load_token_database(
                        self.path, domain=self.domain
                    )
                return database.load_token_database(self.path)
            except FileNotFoundError:
                return database.load_token_database()

    def __init__(
        self, *paths_or_files: _PathOrStr, min_poll_period_s: float = 1.0
    ) -> None:
        self.paths = tuple(self._DatabasePath(path) for path in paths_or_files)
        self.min_poll_period_s = min_poll_period_s
        self._last_checked_time: float = time.time()
        super().__init__(*(path.load() for path in self.paths))

    def _reload_if_changed(self) -> None:
        if time.time() - self._last_checked_time >= self.min_poll_period_s:
            self._last_checked_time = time.time()

            if any(path.updated() for path in self.paths):
                _LOG.info('Changes detected; reloading token database')
                self._initialize_database(path.load() for path in self.paths)

    def lookup(self, token: int) -> List[_TokenizedFormatString]:
        self._reload_if_changed()
        return super().lookup(token)


class PrefixedMessageDecoder:
    """Parses messages that start with a prefix character from a byte stream."""

    def __init__(self, prefix: Union[str, bytes], chars: Union[str, bytes]):
        """Parses prefixed messages.

        Args:
          prefix: one character that signifies the start of a message
          chars: characters allowed in a message
        """
        self._prefix = prefix.encode() if isinstance(prefix, str) else prefix

        if isinstance(chars, str):
            chars = chars.encode()

        # Store the valid message bytes as a set of binary strings.
        self._message_bytes = frozenset(
            chars[i : i + 1] for i in range(len(chars))
        )

        if len(self._prefix) != 1 or self._prefix in self._message_bytes:
            raise ValueError(
                'Invalid prefix {!r}: the prefix must be a single '
                'character that is not a valid message character.'.format(
                    prefix
                )
            )

        self.data = bytearray()

    def _read_next(self, fd: _RawIO) -> Tuple[bytes, int]:
        """Returns the next character and its index."""
        char = fd.read(1) or b''
        index = len(self.data)
        self.data += char
        return char, index

    def read_messages(self, binary_fd: _RawIO) -> Iterator[Tuple[bool, bytes]]:
        """Parses prefixed messages; yields (is_message, contents) chunks."""
        message_start = None

        while True:
            # This reads the file character-by-character. Non-message characters
            # are yielded right away; message characters are grouped.
            char, index = self._read_next(binary_fd)

            # If in a message, keep reading until the message completes.
            if message_start is not None:
                if char in self._message_bytes:
                    continue

                yield True, self.data[message_start:index]
                message_start = None

            # Handle a non-message character.
            if not char:
                return

            if char == self._prefix:
                message_start = index
            else:
                yield False, char

    def transform(
        self, binary_fd: _RawIO, transform: Callable[[bytes], bytes]
    ) -> Iterator[bytes]:
        """Yields the file with a transformation applied to the messages."""
        for is_message, chunk in self.read_messages(binary_fd):
            yield transform(chunk) if is_message else chunk


def _base64_message_regex(prefix: bytes) -> Pattern[bytes]:
    """Returns a regular expression for prefixed base64 tokenized strings."""
    return re.compile(
        # Base64 tokenized strings start with the prefix character ($)
        re.escape(prefix)
        + (
            # Tokenized strings contain 0 or more blocks of four Base64 chars.
            br'(?:[A-Za-z0-9+/\-_]{4})*'
            # The last block of 4 chars may have one or two padding chars (=).
            br'(?:[A-Za-z0-9+/\-_]{3}=|[A-Za-z0-9+/\-_]{2}==)?'
        )
    )


# TODO(hepler): Remove this unnecessary function.
def detokenize_base64(
    detokenizer: Detokenizer,
    data: bytes,
    prefix: Union[str, bytes] = BASE64_PREFIX,
    recursion: int = DEFAULT_RECURSION,
) -> bytes:
    """Alias for detokenizer.detokenize_base64 for backwards compatibility."""
    return detokenizer.detokenize_base64(data, prefix, recursion)


def _follow_and_detokenize_file(
    detokenizer: Detokenizer,
    file: BinaryIO,
    output: BinaryIO,
    prefix: Union[str, bytes],
    poll_period_s: float = 0.01,
) -> None:
    """Polls a file to detokenize it and any appended data."""

    try:
        while True:
            data = file.read()
            if data:
                detokenizer.detokenize_base64_to_file(data, output, prefix)
                output.flush()
            else:
                time.sleep(poll_period_s)
    except KeyboardInterrupt:
        pass


def _handle_base64(
    databases,
    input_file: BinaryIO,
    output: BinaryIO,
    prefix: str,
    show_errors: bool,
    follow: bool,
) -> None:
    """Handles the base64 command line option."""
    # argparse.FileType doesn't correctly handle - for binary files.
    if input_file is sys.stdin:
        input_file = sys.stdin.buffer

    if output is sys.stdout:
        output = sys.stdout.buffer

    detokenizer = Detokenizer(
        tokens.Database.merged(*databases), show_errors=show_errors
    )

    if follow:
        _follow_and_detokenize_file(detokenizer, input_file, output, prefix)
    elif input_file.seekable():
        # Process seekable files all at once, which is MUCH faster.
        detokenizer.detokenize_base64_to_file(input_file.read(), output, prefix)
    else:
        # For non-seekable inputs (e.g. pipes), read one character at a time.
        detokenizer.detokenize_base64_live(input_file, output, prefix)


def _parse_args() -> argparse.Namespace:
    """Parses and return command line arguments."""

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.set_defaults(handler=lambda **_: parser.print_help())

    subparsers = parser.add_subparsers(help='Encoding of the input.')

    base64_help = 'Detokenize Base64-encoded data from a file or stdin.'
    subparser = subparsers.add_parser(
        'base64',
        description=base64_help,
        parents=[database.token_databases_parser()],
        help=base64_help,
    )
    subparser.set_defaults(handler=_handle_base64)
    subparser.add_argument(
        '-i',
        '--input',
        dest='input_file',
        type=argparse.FileType('rb'),
        default=sys.stdin.buffer,
        help='The file from which to read; provide - or omit for stdin.',
    )
    subparser.add_argument(
        '-f',
        '--follow',
        action='store_true',
        help=(
            'Detokenize data appended to input_file as it grows; similar to '
            'tail -f.'
        ),
    )
    subparser.add_argument(
        '-o',
        '--output',
        type=argparse.FileType('wb'),
        default=sys.stdout.buffer,
        help=(
            'The file to which to write the output; '
            'provide - or omit for stdout.'
        ),
    )
    subparser.add_argument(
        '-p',
        '--prefix',
        default=BASE64_PREFIX,
        help=(
            'The one-character prefix that signals the start of a '
            'Base64-encoded message. (default: $)'
        ),
    )
    subparser.add_argument(
        '-s',
        '--show_errors',
        action='store_true',
        help=(
            'Show error messages instead of conversion specifiers when '
            'arguments cannot be decoded.'
        ),
    )

    return parser.parse_args()


def main() -> int:
    args = _parse_args()

    handler = args.handler
    del args.handler

    handler(**vars(args))
    return 0


if __name__ == '__main__':
    if sys.version_info[0] < 3:
        sys.exit('ERROR: The detokenizer command line tools require Python 3.')
    sys.exit(main())



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Console ProgressBar implementation.

Designed to be embedded in an existing prompt_toolkit full screen
application."""

import functools
from typing import (
    Iterable,
    List,
    Optional,
    Sequence,
)

from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.layout import (
    ConditionalContainer,
    FormattedTextControl,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.dimension import AnyDimension, D
from prompt_toolkit.styles import BaseStyle

from prompt_toolkit.shortcuts.progress_bar import (
    ProgressBar,
    ProgressBarCounter,
)
from prompt_toolkit.shortcuts.progress_bar.base import _ProgressControl
from prompt_toolkit.shortcuts.progress_bar.formatters import (
    Formatter,
    IterationsPerSecond,
    Text,
    TimeLeft,
    create_default_formatters,
)


class TextIfNotHidden(Text):
    def format(
        self,
        progress_bar: ProgressBar,
        progress: 'ProgressBarCounter[object]',
        width: int,
    ) -> AnyFormattedText:
        formatted_text = super().format(progress_bar, progress, width)
        if hasattr(progress, 'hide_eta') and progress.hide_eta:  # type: ignore
            formatted_text = [('', ' ' * width)]
        return formatted_text


class IterationsPerSecondIfNotHidden(IterationsPerSecond):
    def format(
        self,
        progress_bar: ProgressBar,
        progress: 'ProgressBarCounter[object]',
        width: int,
    ) -> AnyFormattedText:
        formatted_text = super().format(progress_bar, progress, width)
        if hasattr(progress, 'hide_eta') and progress.hide_eta:  # type: ignore
            formatted_text = [('class:iterations-per-second', ' ' * width)]
        return formatted_text


class TimeLeftIfNotHidden(TimeLeft):
    def format(
        self,
        progress_bar: ProgressBar,
        progress: 'ProgressBarCounter[object]',
        width: int,
    ) -> AnyFormattedText:
        formatted_text = super().format(progress_bar, progress, width)
        if hasattr(progress, 'hide_eta') and progress.hide_eta:  # type: ignore
            formatted_text = [('class:time-left', ' ' * width)]
        return formatted_text


class ProgressBarImpl:
    """ProgressBar for rendering in an existing prompt_toolkit application."""

    def __init__(
        self,
        title: AnyFormattedText = None,
        formatters: Optional[Sequence[Formatter]] = None,
        style: Optional[BaseStyle] = None,
    ) -> None:
        self.title = title
        self.formatters = formatters or create_default_formatters()
        self.counters: List[ProgressBarCounter[object]] = []
        self.style = style

        # Create UI Application.
        title_toolbar = ConditionalContainer(
            Window(
                FormattedTextControl(lambda: self.title),
                height=1,
                style='class:progressbar,title',
            ),
            filter=Condition(lambda: self.title is not None),
        )

        def width_for_formatter(formatter: Formatter) -> AnyDimension:
            # Needs to be passed as callable (partial) to the 'width'
            # parameter, because we want to call it on every resize.
            return formatter.get_width(progress_bar=self)  # type: ignore

        progress_controls = [
            Window(
                content=_ProgressControl(self, f, None),  # type: ignore
                width=functools.partial(width_for_formatter, f),
            )
            for f in self.formatters
        ]

        self.container = HSplit(
            [
                title_toolbar,
                VSplit(
                    progress_controls,
                    height=lambda: D(
                        min=len(self.counters), max=len(self.counters)
                    ),
                ),
            ]
        )

    def __pt_container__(self):
        return self.container

    def __exit__(self, *a: object) -> None:
        pass

    def __call__(
        self,
        data: Optional[Iterable] = None,
        label: AnyFormattedText = '',
        remove_when_done: bool = False,
        total: Optional[int] = None,
    ) -> 'ProgressBarCounter':
        """
        Start a new counter.

        :param label: Title text or description for this progress. (This can be
            formatted text as well).
        :param remove_when_done: When `True`, hide this progress bar.
        :param total: Specify the maximum value if it can't be calculated by
            calling ``len``.
        """
        counter = ProgressBarCounter(
            self,  # type: ignore
            data,
            label=label,
            remove_when_done=remove_when_done,
            total=total,
        )
        self.counters.append(counter)
        return counter



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Reads and parses the Pigweed config file.

See also https://pigweed.dev/seed/0101-pigweed.json.html.
"""

import json
import os


def _get_project_root(env=None):
    if not env:
        env = os.environ
    for var in ('PW_PROJECT_ROOT', 'PW_ROOT'):
        if var in env:
            return env[var]
    raise ValueError('environment variable PW_PROJECT_ROOT not set')


def path(env=None):
    """Return the path where pigweed.json should reside."""
    return os.path.join(_get_project_root(env=env), 'pigweed.json')


def load(env=None):
    """Load pigweed.json if it exists and return the contents."""
    config_path = path(env=env)
    if not os.path.isfile(config_path):
        return {}
    with open(config_path, 'r') as ins:
        return json.load(ins)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for using HDLC with pw_rpc."""

from concurrent.futures import ThreadPoolExecutor
import io
import logging
from queue import SimpleQueue
import sys
import threading
import time
import socket
import subprocess
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterable,
    List,
    NoReturn,
    Optional,
    Sequence,
    Union,
)

from pw_protobuf_compiler import python_protos
import pw_rpc
from pw_rpc import callback_client

from pw_hdlc.decode import Frame, FrameDecoder
from pw_hdlc import encode

_LOG = logging.getLogger(__name__)

STDOUT_ADDRESS = 1
DEFAULT_ADDRESS = ord('R')
_VERBOSE = logging.DEBUG - 1


def channel_output(
    writer: Callable[[bytes], Any],
    address: int = DEFAULT_ADDRESS,
    delay_s: float = 0,
) -> Callable[[bytes], None]:
    """Returns a function that can be used as a channel output for pw_rpc."""

    if delay_s:

        def slow_write(data: bytes) -> None:
            """Slows down writes in case unbuffered serial is in use."""
            for byte in data:
                time.sleep(delay_s)
                writer(bytes([byte]))

        return lambda data: slow_write(encode.ui_frame(address, data))

    def write_hdlc(data: bytes):
        frame = encode.ui_frame(address, data)
        _LOG.log(_VERBOSE, 'Write %2d B: %s', len(frame), frame)
        writer(frame)

    return write_hdlc


def _handle_error(frame: Frame) -> None:
    _LOG.error('Failed to parse frame: %s', frame.status.value)
    _LOG.debug('%s', frame.data)


FrameHandlers = Dict[int, Callable[[Frame], Any]]


def read_and_process_data(
    read: Callable[[], bytes],
    on_read_error: Callable[[Exception], Any],
    frame_handlers: FrameHandlers,
    error_handler: Callable[[Frame], Any] = _handle_error,
    handler_threads: Optional[int] = 1,
) -> NoReturn:
    """Continuously reads and handles HDLC frames.

    Passes frames to an executor that calls frame handler functions in other
    threads.
    """

    def handle_frame(frame: Frame):
        try:
            if not frame.ok():
                error_handler(frame)
                return

            try:
                frame_handlers[frame.address](frame)
            except KeyError:
                _LOG.warning(
                    'Unhandled frame for address %d: %s', frame.address, frame
                )
        except:  # pylint: disable=bare-except
            _LOG.exception('Exception in HDLC frame handler thread')

    decoder = FrameDecoder()

    # Execute callbacks in a ThreadPoolExecutor to decouple reading the input
    # stream from handling the data. That way, if a handler function takes a
    # long time or crashes, this reading thread is not interrupted.
    with ThreadPoolExecutor(max_workers=handler_threads) as executor:
        while True:
            try:
                data = read()
            except Exception as exc:  # pylint: disable=broad-except
                on_read_error(exc)
                continue

            if data:
                _LOG.log(_VERBOSE, 'Read %2d B: %s', len(data), data)

                for frame in decoder.process_valid_frames(data):
                    executor.submit(handle_frame, frame)


def write_to_file(data: bytes, output: BinaryIO = sys.stdout.buffer):
    output.write(data + b'\n')
    output.flush()


def default_channels(write: Callable[[bytes], Any]) -> List[pw_rpc.Channel]:
    return [pw_rpc.Channel(1, channel_output(write))]


PathsModulesOrProtoLibrary = Union[
    Iterable[python_protos.PathOrModule], python_protos.Library
]


class HdlcRpcClient:
    """An RPC client configured to run over HDLC."""

    def __init__(
        self,
        read: Callable[[], bytes],
        paths_or_modules: PathsModulesOrProtoLibrary,
        channels: Iterable[pw_rpc.Channel],
        output: Callable[[bytes], Any] = write_to_file,
        client_impl: Optional[pw_rpc.client.ClientImpl] = None,
        *,
        _incoming_packet_filter_for_testing: Optional[
            pw_rpc.ChannelManipulator
        ] = None,
    ):
        """Creates an RPC client configured to communicate using HDLC.

        Args:
          read: Function that reads bytes; e.g serial_device.read.
          paths_or_modules: paths to .proto files or proto modules
          channel: RPC channels to use for output
          output: where to write "stdout" output from the device
        """
        if isinstance(paths_or_modules, python_protos.Library):
            self.protos = paths_or_modules
        else:
            self.protos = python_protos.Library.from_paths(paths_or_modules)

        if client_impl is None:
            client_impl = callback_client.Impl()

        self.client = pw_rpc.Client.from_modules(
            client_impl, channels, self.protos.modules()
        )

        rpc_output: Callable[[bytes], Any] = self._handle_rpc_packet
        if _incoming_packet_filter_for_testing is not None:
            _incoming_packet_filter_for_testing.send_packet = rpc_output
            rpc_output = _incoming_packet_filter_for_testing

        frame_handlers: FrameHandlers = {
            DEFAULT_ADDRESS: lambda frame: rpc_output(frame.data),
            STDOUT_ADDRESS: lambda frame: output(frame.data),
        }

        # Start background thread that reads and processes RPC packets.
        threading.Thread(
            target=read_and_process_data,
            daemon=True,
            args=(read, lambda exc: None, frame_handlers),
        ).start()

    def rpcs(self, channel_id: Optional[int] = None) -> Any:
        """Returns object for accessing services on the specified channel.

        This skips some intermediate layers to make it simpler to invoke RPCs
        from an HdlcRpcClient. If only one channel is in use, the channel ID is
        not necessary.
        """
        if channel_id is None:
            return next(iter(self.client.channels())).rpcs

        return self.client.channel(channel_id).rpcs

    def _handle_rpc_packet(self, packet: bytes) -> None:
        if not self.client.process_packet(packet):
            _LOG.error('Packet not handled by RPC client: %s', packet)


def _try_connect(port: int, attempts: int = 10) -> socket.socket:
    """Tries to connect to the specified port up to the given number of times.

    This is helpful when connecting to a process that was started by this
    script. The process may need time to start listening for connections, and
    that length of time can vary. This retries with a short delay rather than
    having to wait for the worst case delay every time.
    """
    timeout_s = 0.001
    while True:
        time.sleep(timeout_s)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(('localhost', port))
            return sock
        except ConnectionRefusedError:
            sock.close()
            attempts -= 1
            if attempts <= 0:
                raise

            timeout_s *= 2


class SocketSubprocess:
    """Executes a subprocess and connects to it with a socket."""

    def __init__(self, command: Sequence, port: int) -> None:
        self._server_process = subprocess.Popen(command, stdin=subprocess.PIPE)
        self.stdin = self._server_process.stdin

        try:
            self.socket: socket.socket = _try_connect(port)  # 🧦
        except:
            self._server_process.terminate()
            self._server_process.communicate()
            raise

    def close(self) -> None:
        try:
            self.socket.close()
        finally:
            self._server_process.terminate()
            self._server_process.communicate()

    def __enter__(self) -> 'SocketSubprocess':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()


class HdlcRpcLocalServerAndClient:
    """Runs an RPC server in a subprocess and connects to it over a socket.

    This can be used to run a local RPC server in an integration test.
    """

    def __init__(
        self,
        server_command: Sequence,
        port: int,
        protos: PathsModulesOrProtoLibrary,
        *,
        incoming_processor: Optional[pw_rpc.ChannelManipulator] = None,
        outgoing_processor: Optional[pw_rpc.ChannelManipulator] = None,
    ) -> None:
        """Creates a new HdlcRpcLocalServerAndClient."""

        self.server = SocketSubprocess(server_command, port)

        self._bytes_queue: 'SimpleQueue[bytes]' = SimpleQueue()
        self._read_thread = threading.Thread(target=self._read_from_socket)
        self._read_thread.start()

        self.output = io.BytesIO()

        self.channel_output: Any = self.server.socket.sendall

        self._incoming_processor = incoming_processor
        if outgoing_processor is not None:
            outgoing_processor.send_packet = self.channel_output
            self.channel_output = outgoing_processor

        self.client = HdlcRpcClient(
            self._bytes_queue.get,
            protos,
            default_channels(self.channel_output),
            self.output.write,
            _incoming_packet_filter_for_testing=incoming_processor,
        ).client

    def _read_from_socket(self):
        while True:
            data = self.server.socket.recv(4096)
            self._bytes_queue.put(data)
            if not data:
                return

    def close(self):
        self.server.close()
        self.output.close()
        self._read_thread.join()

    def __enter__(self) -> 'HdlcRpcLocalServerAndClient':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_ide CLI command handlers."""

import logging
from pathlib import Path
import shlex
import shutil
import subprocess
import sys
from typing import cast, Callable, List, Optional, Set, Tuple, Union

from pw_cli.color import colors

from pw_ide.cpp import (
    ClangdSettings,
    compdb_generate_file_path,
    CppCompilationDatabase,
    CppCompilationDatabasesMap,
    CppIdeFeaturesState,
    delete_compilation_databases,
    delete_compilation_database_caches,
    MAX_COMMANDS_TARGET_FILENAME,
)

from pw_ide.exceptions import (
    BadCompDbException,
    InvalidTargetException,
    MissingCompDbException,
)

from pw_ide.python import PythonPaths

from pw_ide.settings import (
    PigweedIdeSettings,
    SupportedEditor,
    SupportedEditorName,
)

from pw_ide import vscode
from pw_ide.vscode import VscSettingsManager, VscSettingsType


def _no_color(msg: str) -> str:
    return msg


def _split_lines(msg: Union[str, List[str]]) -> Tuple[str, List[str]]:
    """Turn a list of strings into a tuple of the first and list of rest."""
    if isinstance(msg, str):
        return (msg, [])

    return (msg[0], msg[1:])


class StatusReporter:
    """Print user-friendly status reports to the terminal for CLI tools.

    The output of ``demo()`` looks something like this, but more colorful:

    .. code-block:: none

       • FYI, here's some information:
         Lorem ipsum dolor sit amet, consectetur adipiscing elit.
         Donec condimentum metus molestie metus maximus ultricies ac id dolor.
       ✓ This is okay, no changes needed.
       ✓ We changed some things successfully!
       ⚠ Uh oh, you might want to be aware of this.
       ❌ This is bad! Things might be broken!

    You can instead redirect these lines to logs without formatting by
    substituting ``LoggingStatusReporter``. Consumers of this should be
    designed to take any subclass and not make assumptions about where the
    output will go. But the reason you would choose this over plain logging is
    because you want to support pretty-printing to the terminal.

    This is also "themable" in the sense that you can subclass this, override
    the methods with whatever formatting you want, and supply the subclass to
    anything that expects an instance of this.

    Key:

    - info: Plain ol' informational status.
    - ok: Something was checked and it was okay.
    - new: Something needed to be changed/updated and it was successfully.
    - wrn: Warning, non-critical.
    - err: Error, critical.

    This doesn't expose the %-style string formatting that is used in idiomatic
    Python logging, but this shouldn't be used for performance-critical logging
    situations anyway.
    """

    def _report(  # pylint: disable=no-self-use
        self,
        msg: Union[str, List[str]],
        color: Callable[[str], str],
        char: str,
        func: Callable,
        silent: bool,
    ) -> None:
        """Actually print/log/whatever the status lines."""
        first_line, rest_lines = _split_lines(msg)
        first_line = color(f'{char} {first_line}')
        spaces = ' ' * len(char)
        rest_lines = [color(f'{spaces} {line}') for line in rest_lines]

        if not silent:
            for line in [first_line, *rest_lines]:
                func(line)

    def demo(self):
        """Run this to see what your status reporter output looks like."""
        self.info(
            [
                'FYI, here\'s some information:',
                'Lorem ipsum dolor sit amet, consectetur adipiscing elit.',
                'Donec condimentum metus molestie metus maximus ultricies '
                'ac id dolor.',
            ]
        )
        self.ok('This is okay, no changes needed.')
        self.new('We changed some things successfully!')
        self.wrn('Uh oh, you might want to be aware of this.')
        self.err('This is bad! Things might be broken!')

    def info(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, _no_color, '\u2022', print, silent)

    def ok(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, colors().blue, '\u2713', print, silent)

    def new(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, colors().green, '\u2713', print, silent)

    def wrn(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, colors().yellow, '\u26A0', print, silent)

    def err(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, colors().red, '\u274C', print, silent)


class LoggingStatusReporter(StatusReporter):
    """Print status lines to logs instead of to the terminal."""

    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger
        super().__init__()

    def _report(
        self,
        msg: Union[str, List[str]],
        color: Callable[[str], str],
        char: str,
        func: Callable,
        silent: bool,
    ) -> None:
        first_line, rest_lines = _split_lines(msg)

        if not silent:
            for line in [first_line, *rest_lines]:
                func(line)

    def info(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.info, silent)

    def ok(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.info, silent)

    def new(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.info, silent)

    def wrn(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.warning, silent)

    def err(self, msg: Union[str, List[str]], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.error, silent)


def _make_working_dir(
    reporter: StatusReporter, settings: PigweedIdeSettings, quiet: bool = False
) -> None:
    if not settings.working_dir.exists():
        settings.working_dir.mkdir()
        reporter.new(
            'Initialized the Pigweed IDE working directory at '
            f'{settings.working_dir}'
        )
    else:
        if not quiet:
            reporter.ok(
                'Pigweed IDE working directory already present at '
                f'{settings.working_dir}'
            )


def _report_unrecognized_editor(reporter: StatusReporter, editor: str) -> None:
    supported_editors = ', '.join(sorted([ed.value for ed in SupportedEditor]))
    reporter.wrn(f'Unrecognized editor: {editor}')
    reporter.wrn('This may not be an automatically-supported editor.')
    reporter.wrn(f'Automatically-supported editors: {supported_editors}')


def cmd_clear(
    compdb: bool,
    cache: bool,
    editor: Optional[SupportedEditorName],
    editor_backups: Optional[SupportedEditorName],
    silent: bool = False,
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Clear components of the IDE features.

    In contrast to the ``reset`` subcommand, ``clear`` allows you to specify
    components to delete. You will not need this command under normal
    circumstances.
    """
    if compdb:
        delete_compilation_databases(pw_ide_settings)
        reporter.wrn('Cleared compilation databases', silent)

    if cache:
        delete_compilation_database_caches(pw_ide_settings)
        reporter.wrn('Cleared compilation database caches', silent)

    if editor is not None:
        try:
            validated_editor = SupportedEditor(editor)
        except ValueError:
            _report_unrecognized_editor(reporter, cast(str, editor))
            sys.exit(1)

        if validated_editor == SupportedEditor.VSCODE:
            vsc_settings_manager = VscSettingsManager(pw_ide_settings)
            vsc_settings_manager.delete_all_active_settings()

        reporter.wrn(
            f'Cleared active settings for {validated_editor.value}', silent
        )

    if editor_backups is not None:
        try:
            validated_editor = SupportedEditor(editor_backups)
        except ValueError:
            _report_unrecognized_editor(reporter, cast(str, editor))
            sys.exit(1)

        if validated_editor == SupportedEditor.VSCODE:
            vsc_settings_manager = VscSettingsManager(pw_ide_settings)
            vsc_settings_manager.delete_all_backups()

        reporter.wrn(
            f'Cleared backup settings for {validated_editor.value}',
            silent=silent,
        )


def cmd_reset(
    hard: bool = False,
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Reset IDE settings.

    This will clear your .pw_ide working directory and active settings for
    supported editors, restoring your repository to a pre-"pw ide setup" state.
    Any clangd caches in the working directory will not be removed, so that they
    don't need to be generated again later. All backed up supported editor
    settings will also be left in place.

    Adding the --hard flag will completely delete the .pw_ide directory and all
    supported editor backup settings, restoring your repository to a
    pre-`pw ide setup` state.

    This command does not affect this project's pw_ide and editor settings or
    your own pw_ide and editor override settings.
    """
    delete_compilation_databases(pw_ide_settings)
    vsc_settings_manager = VscSettingsManager(pw_ide_settings)
    vsc_settings_manager.delete_all_active_settings()

    if hard:
        try:
            shutil.rmtree(pw_ide_settings.working_dir)
        except FileNotFoundError:
            pass

        vsc_settings_manager.delete_all_backups()

    reporter.wrn('Pigweed IDE settings were reset!')


def cmd_setup(
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Set up or update your Pigweed project IDE features.

    This will automatically set up your development environment with all the
    features that Pigweed IDE supports, with sensible defaults.

    At minimum, this command will create the .pw_ide working directory and
    create settings files for all supported editors. Projects can define
    additional setup steps in .pw_ide.yaml.

    When new IDE features are introduced in the future (either by Pigweed or
    your downstream project), you can re-run this command to set up the new
    features. It will not overwrite or break any of your existing configuration.
    """
    _make_working_dir(reporter, pw_ide_settings)

    if pw_ide_settings.editor_enabled('vscode'):
        cmd_vscode()

    for command in pw_ide_settings.setup:
        subprocess.run(shlex.split(command))


def cmd_vscode(
    include: Optional[List[VscSettingsType]] = None,
    exclude: Optional[List[VscSettingsType]] = None,
    no_override: bool = False,
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Configure support for Visual Studio Code.

    This will replace your current Visual Studio Code (VSC) settings for this
    project (in ``.vscode/settings.json``, etc.) with the following sets of
    settings, in order:

    - The Pigweed default settings
    - Your project's settings, if any (in ``.vscode/pw_project_settings.json``)
    - Your personal settings, if any (in ``.vscode/pw_user_settings.json``)

    In other words, settings files lower on the list can override settings
    defined by those higher on the list. Settings defined in the sources above
    are not active in VSC until they are merged and output to the current
    settings file by running:

    .. code-block:: bash

       pw ide vscode

    Refer to the Visual Studio Code documentation for more information about
    these settings: https://code.visualstudio.com/docs/getstarted/settings

    This command also manages VSC tasks (``.vscode/tasks.json``) and extensions
    (``.vscode/extensions.json``). You can explicitly control which of these
    settings types ("settings", "tasks", and "extensions") is modified by
    this command by using the ``--include`` or ``--exclude`` options.

    Your current VSC settings will never change unless you run ``pw ide``
    commands. Since the current VSC settings are an artifact built from the
    three settings files described above, you should avoid manually editing
    that file; it will be replaced the next time you run ``pw ide vscode``. A
    backup of your previous settings file will be made, and you can diff it
    against the new file to see what changed.

    These commands will never modify your VSC user settings, which are
    stored outside of the project repository and apply globally to all VSC
    instances.

    The settings files are system-specific and shouldn't be checked into the
    repository, except for the project settings (those with ``pw_project_``),
    which can be used to define consistent settings for everyone working on the
    project.

    Note that support for VSC can be disabled at the project level or the user
    level by adding the following to .pw_ide.yaml or .pw_ide.user.yaml
    respectively:

    .. code-block:: yaml

        editors:
          vscode: false

    Likewise, it can be enabled by setting that value to true. It is enabled by
    default.
    """
    if not pw_ide_settings.editor_enabled('vscode'):
        reporter.wrn('Visual Studio Code support is disabled in settings!')
        sys.exit(1)

    if not vscode.DEFAULT_SETTINGS_PATH.exists():
        vscode.DEFAULT_SETTINGS_PATH.mkdir()

    vsc_manager = VscSettingsManager(pw_ide_settings)

    if include is None:
        include_set = set(VscSettingsType.all())
    else:
        include_set = set(include)

    if exclude is None:
        exclude_set: Set[VscSettingsType] = set()
    else:
        exclude_set = set(exclude)

    types_to_update = cast(
        List[VscSettingsType], tuple(include_set - exclude_set)
    )

    for settings_type in types_to_update:
        active_settings_existed = vsc_manager.active(settings_type).is_present()

        if no_override and active_settings_existed:
            reporter.ok(
                f'Visual Studio Code active {settings_type.value} '
                'already present; will not overwrite'
            )

        else:
            with vsc_manager.active(settings_type).modify(
                reinit=True
            ) as active_settings:
                vsc_manager.default(settings_type).sync_to(active_settings)
                vsc_manager.project(settings_type).sync_to(active_settings)
                vsc_manager.user(settings_type).sync_to(active_settings)

            verb = 'Updated' if active_settings_existed else 'Created'
            reporter.new(
                f'{verb} Visual Studio Code active ' f'{settings_type.value}'
            )


# TODO(chadnorvell): Break up this function.
# The linting errors are a nuisance but they're beginning to have a point.
def cmd_cpp(  # pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements
    should_list_targets: bool,
    should_get_target: bool,
    target_to_set: Optional[str],
    compdb_file_paths: Optional[List[Path]],
    build_dir: Optional[Path],
    use_default_target: bool = False,
    should_run_ninja: bool = False,
    should_run_gn: bool = False,
    override_current_target: bool = True,
    clangd_command: bool = False,
    clangd_command_system: Optional[str] = None,
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Configure C/C++ code intelligence support.

    Code intelligence can be provided by clangd or other language servers that
    use the clangd compilation database format, defined at:
    https://clang.llvm.org/docs/JSONCompilationDatabase.html

    This command helps you use clangd with Pigweed projects, which use multiple
    toolchains within a distinct environment, and often define multiple targets.
    This means compilation units are likely have multiple compile commands, and
    clangd is not equipped to deal with this out of the box. We handle this by:

    - Processing the compilation database produced the build system into
      multiple internally-consistent compilation databases, one for each target
      (where a "target" is a particular build for a particular system using a
      particular toolchain).

    - Providing commands to select which target you want to use for code
      analysis.

    Refer to the Pigweed documentation or your build system's documentation to
    learn how to produce a clangd compilation database. Once you have one, run
    this command to process it (or provide a glob to process multiple):

    .. code-block:: bash

        pw ide cpp --process {path to compile_commands.json}

    If you're using GN to generate the compilation database, you can do that and
    process it in a single command:

    .. code-block:: bash

       pw ide cpp --gn

    You can do the same for a Ninja build (whether it was generated by GN or
    another way):

    .. code-block:: bash

       pw ide cpp --ninja

    You can now examine the targets that are available to you:

    .. code-block:: bash

        pw ide cpp --list

    ... and select the target you want to use:

    .. code-block:: bash

        pw ide cpp --set host_clang

    As long as your editor or language server plugin is properly configured, you
    will now get code intelligence features relevant to that particular target.

    You can see what target is selected by running:

    .. code-block:: bash

        pw ide cpp

    Whenever you switch to a target you haven't used before, clangd will need to
    index the build, which may take several minutes. These indexes are cached,
    so you can switch between targets without re-indexing each time.

    If your build configuration changes significantly (e.g. you add a new file
    to the project), you will need to re-process the compilation database for
    that change to be recognized. Your target selection will not change, and
    your index will only need to be incrementally updated.

    You can generate the clangd command your editor needs to run with:

    .. code-block:: bash

        pw ide cpp --clangd-command

    If your editor uses JSON for configuration, you can export the same command
    in that format:

    .. code-block:: bash

        pw ide cpp --clangd-command-for json
    """
    _make_working_dir(reporter, pw_ide_settings, quiet=True)

    # If true, no arguments were provided so we do the default behavior.
    default = True

    build_dir = (
        build_dir if build_dir is not None else pw_ide_settings.build_dir
    )

    if compdb_file_paths is not None:
        should_process = True

        if len(compdb_file_paths) == 0:
            compdb_file_paths = pw_ide_settings.compdb_paths_expanded
    else:
        should_process = False
        # This simplifies typing in the rest of this method. We rely on
        # `should_process` instead of the status of this variable.
        compdb_file_paths = []

    # Order of operations matters from here on. It should be possible to run
    # a build system command to generate a compilation database, then process
    # the compilation database, then successfully set the target in a single
    # command.

    # Use Ninja to generate the initial compile_commands.json
    if should_run_ninja:
        default = False

        ninja_commands = ['ninja', '-t', 'compdb']
        reporter.info(f'Running Ninja: {" ".join(ninja_commands)}')

        output_compdb_file_path = build_dir / compdb_generate_file_path()

        try:
            # Ninja writes to STDOUT, so we capture to a file.
            with open(output_compdb_file_path, 'w') as compdb_file:
                result = subprocess.run(
                    ninja_commands,
                    cwd=build_dir,
                    stdout=compdb_file,
                    stderr=subprocess.PIPE,
                )
        except FileNotFoundError:
            reporter.err(f'Could not open path! {str(output_compdb_file_path)}')

        if result.returncode == 0:
            reporter.info('Ran Ninja successfully!')
            should_process = True
            compdb_file_paths.append(output_compdb_file_path)
        else:
            reporter.err('Something went wrong!')
            # Convert from bytes and remove trailing newline
            err = result.stderr.decode().split('\n')[:-1]

            for line in err:
                reporter.err(line)

            sys.exit(1)

    # Use GN to generate the initial compile_commands.json
    if should_run_gn:
        default = False

        gn_commands = ['gn', 'gen', str(build_dir), '--export-compile-commands']

        try:
            with open(build_dir / 'args.gn') as args_file:
                gn_args = [
                    line
                    for line in args_file.readlines()
                    if not line.startswith('#')
                ]
        except FileNotFoundError:
            gn_args = []

        gn_args_string = 'none' if len(gn_args) == 0 else ', '.join(gn_args)

        reporter.info(
            [f'Running GN: {" ".join(gn_commands)} (args: {gn_args_string})']
        )

        result = subprocess.run(gn_commands, capture_output=True)
        gn_status_lines = ['Ran GN successfully!']

        if result.returncode == 0:
            # Convert from bytes and remove trailing newline
            out = result.stdout.decode().split('\n')[:-1]

            for line in out:
                gn_status_lines.append(line)

            reporter.info(gn_status_lines)
            should_process = True
            output_compdb_file_path = build_dir / compdb_generate_file_path()
            compdb_file_paths.append(output_compdb_file_path)
        else:
            reporter.err('Something went wrong!')
            # Convert from bytes and remove trailing newline
            err = result.stderr.decode().split('\n')[:-1]

            for line in err:
                reporter.err(line)

            sys.exit(1)

    if should_process:
        default = False
        prev_targets = len(CppIdeFeaturesState(pw_ide_settings))
        compdb_databases: List[CppCompilationDatabasesMap] = []
        last_processed_path = Path()

        for compdb_file_path in compdb_file_paths:
            # If the path is a dir, append the default compile commands
            # file name.
            if compdb_file_path.is_dir():
                compdb_file_path /= compdb_generate_file_path()

            try:
                compdb_databases.append(
                    CppCompilationDatabase.load(
                        Path(compdb_file_path), build_dir
                    ).process(
                        settings=pw_ide_settings,
                        path_globs=pw_ide_settings.clangd_query_drivers(),
                    )
                )
            except MissingCompDbException:
                reporter.err(f'File not found: {str(compdb_file_path)}')

                if '*' in str(compdb_file_path):
                    reporter.wrn(
                        'It looks like you provided a glob that '
                        'did not match any files.'
                    )

                sys.exit(1)
            # TODO(chadnorvell): Recover more gracefully from errors.
            except BadCompDbException:
                reporter.err(
                    'File does not match compilation database format: '
                    f'{str(compdb_file_path)}'
                )
                sys.exit(1)

            last_processed_path = compdb_file_path

        if len(compdb_databases) == 0:
            reporter.err(
                'No compilation databases found in: '
                f'{str(compdb_file_paths)}'
            )
            sys.exit(1)

        try:
            CppCompilationDatabasesMap.merge(*compdb_databases).write()
        except TypeError:
            reporter.err('Could not serialize file to JSON!')

        total_targets = len(CppIdeFeaturesState(pw_ide_settings))
        new_targets = total_targets - prev_targets

        if len(compdb_file_paths) == 1:
            processed_text = str(last_processed_path)
        else:
            processed_text = f'{len(compdb_file_paths)} compilation databases'

        reporter.new(
            [
                f'Processed {processed_text} '
                f'to {pw_ide_settings.working_dir}',
                f'{total_targets} targets are now available '
                f'({new_targets} are new)',
            ]
        )

    if use_default_target:
        defined_default = pw_ide_settings.default_target
        max_commands_target: Optional[str] = None

        try:
            with open(
                pw_ide_settings.working_dir / MAX_COMMANDS_TARGET_FILENAME
            ) as max_commands_target_file:
                max_commands_target = max_commands_target_file.readline()
        except FileNotFoundError:
            pass

        if defined_default is None and max_commands_target is None:
            reporter.err('Can\'t use default target because none is defined!')
            reporter.wrn('Have you processed a compilation database yet?')
            sys.exit(1)

        target_to_set = (
            defined_default
            if defined_default is not None
            else max_commands_target
        )

    if target_to_set is not None:
        default = False

        # Always set the target if it's not already set, but if it is,
        # respect the --no-override flag.
        should_set_target = (
            CppIdeFeaturesState(pw_ide_settings).current_target is None
            or override_current_target
        )

        if should_set_target:
            try:
                CppIdeFeaturesState(
                    pw_ide_settings
                ).current_target = target_to_set
            except InvalidTargetException:
                reporter.err(
                    [
                        f'Invalid target! {target_to_set} not among the '
                        'defined targets.',
                        'Check .pw_ide.yaml or .pw_ide.user.yaml for defined '
                        'targets.',
                    ]
                )
                sys.exit(1)
            except MissingCompDbException:
                reporter.err(
                    [
                        f'File not found for target! {target_to_set}',
                        'Did you run pw ide cpp --process '
                        '{path to compile_commands.json}?',
                    ]
                )
                sys.exit(1)

            reporter.new(
                'Set C/C++ language server analysis target to: '
                f'{target_to_set}'
            )
        else:
            reporter.ok(
                'Target already is set and will not be overridden: '
                f'{CppIdeFeaturesState(pw_ide_settings).current_target}'
            )

    if clangd_command:
        default = False
        reporter.info(
            [
                'Command to run clangd with Pigweed paths:',
                ClangdSettings(pw_ide_settings).command(),
            ]
        )

    if clangd_command_system is not None:
        default = False
        reporter.info(
            [
                'Command to run clangd with Pigweed paths for '
                f'{clangd_command_system}:',
                ClangdSettings(pw_ide_settings).command(clangd_command_system),
            ]
        )

    if should_list_targets:
        default = False
        targets_list_status = [
            'C/C++ targets available for language server analysis:'
        ]

        for target in sorted(
            CppIdeFeaturesState(pw_ide_settings).enabled_available_targets
        ):
            targets_list_status.append(f'\t{target}')

        reporter.info(targets_list_status)

    if should_get_target or default:
        reporter.info(
            'Current C/C++ language server analysis target: '
            f'{CppIdeFeaturesState(pw_ide_settings).current_target}'
        )


def cmd_python(
    should_print_venv: bool, reporter: StatusReporter = StatusReporter()
) -> None:
    """Configure Python code intelligence support.

    You can generate the path to the Python virtual environment interpreter that
    your editor/language server should use with:

    .. code-block:: bash

       pw ide python --venv
    """
    # If true, no arguments were provided and we should do the default
    # behavior.
    default = True

    if should_print_venv or default:
        reporter.info(
            [
                'Location of the Pigweed Python virtual environment:',
                PythonPaths().interpreter,
            ]
        )



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Protocol version-aware chunk message wrapper."""

import enum
from typing import Any, Optional

from pw_status import Status

try:
    from pw_transfer import transfer_pb2
except ImportError:
    # For the bazel build, which puts generated protos in a different location.
    from pigweed.pw_transfer import transfer_pb2  # type: ignore


class ProtocolVersion(enum.Enum):
    """Supported versions of pw_transfer's RPC data transfer protocol."""

    # Protocol version not known or not set.
    UNKNOWN = 0

    # The original transfer protocol, prior to transfer start/end handshakes.
    LEGACY = 1

    # Second version of the transfer protocol. Guarantees type fields on all
    # chunks, deprecates pending_bytes in favor of window_end_offset, splits
    # transfer resource IDs from ephemeral session IDs, and adds a handshake
    # to the start and end of all transfer sessions.
    VERSION_TWO = 2

    # Alias to the most up-to-date version of the transfer protocol.
    LATEST = VERSION_TWO


_ChunkType = transfer_pb2.Chunk.Type


class Chunk:
    """A chunk exchanged in a pw_transfer stream.

    Wraps the generated protobuf Chunk class with protocol-aware field encoding
    and decoding.
    """

    Type = transfer_pb2.Chunk.Type

    # TODO(frolv): Figure out how to make the chunk type annotation work.
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        protocol_version: ProtocolVersion,
        chunk_type: Any,
        session_id: int = 0,
        resource_id: Optional[int] = None,
        offset: int = 0,
        window_end_offset: int = 0,
        data: bytes = b'',
        remaining_bytes: Optional[int] = None,
        max_chunk_size_bytes: Optional[int] = None,
        min_delay_microseconds: Optional[int] = None,
        status: Optional[Status] = None,
    ):
        self.protocol_version = protocol_version
        self.type = chunk_type
        self.session_id = session_id
        self.resource_id = resource_id
        self.offset = offset
        self.window_end_offset = window_end_offset
        self.data = data
        self.remaining_bytes = remaining_bytes
        self.max_chunk_size_bytes = max_chunk_size_bytes
        self.min_delay_microseconds = min_delay_microseconds
        self.status = status

    @classmethod
    def from_message(cls, message: transfer_pb2.Chunk) -> 'Chunk':
        """Parses a Chunk from a protobuf message."""

        # Some very old versions of transfer don't always encode chunk types,
        # so they must be deduced.
        #
        # The type-less legacy transfer protocol doesn't support handshakes or
        # continuation parameters. Therefore, there are only three possible
        # types: start, data, and retransmit.
        if message.HasField('type'):
            chunk_type = message.type
        elif (
            message.offset == 0
            and not message.data
            and not message.HasField('status')
        ):
            chunk_type = Chunk.Type.START
        elif message.data:
            chunk_type = Chunk.Type.DATA
        else:
            chunk_type = Chunk.Type.PARAMETERS_RETRANSMIT

        chunk = cls(
            ProtocolVersion.UNKNOWN,
            chunk_type,
            offset=message.offset,
            window_end_offset=message.window_end_offset,
            data=message.data,
        )

        if message.HasField('session_id'):
            chunk.protocol_version = ProtocolVersion.VERSION_TWO
            chunk.session_id = message.session_id
        else:
            chunk.protocol_version = ProtocolVersion.LEGACY
            chunk.session_id = message.transfer_id

        if message.HasField('resource_id'):
            chunk.resource_id = message.resource_id

        if message.HasField('protocol_version'):
            # An explicitly specified protocol version overrides any inferred
            # one.
            chunk.protocol_version = ProtocolVersion(message.protocol_version)

        if message.HasField('pending_bytes'):
            chunk.window_end_offset = message.offset + message.pending_bytes

        if message.HasField('remaining_bytes'):
            chunk.remaining_bytes = message.remaining_bytes

        if message.HasField('max_chunk_size_bytes'):
            chunk.max_chunk_size_bytes = message.max_chunk_size_bytes

        if message.HasField('min_delay_microseconds'):
            chunk.min_delay_microseconds = message.min_delay_microseconds

        if message.HasField('status'):
            chunk.status = Status(message.status)

        if chunk.protocol_version is ProtocolVersion.UNKNOWN:
            # If no fields in the chunk specified its protocol version,
            # assume it is a legacy chunk.
            chunk.protocol_version = ProtocolVersion.LEGACY

        return chunk

    def to_message(self) -> transfer_pb2.Chunk:
        """Converts the chunk to a protobuf message."""
        message = transfer_pb2.Chunk(
            offset=self.offset,
            window_end_offset=self.window_end_offset,
            type=self.type,
        )

        if self.resource_id is not None:
            message.resource_id = self.resource_id

        if self.protocol_version is ProtocolVersion.VERSION_TWO:
            if self.session_id != 0:
                message.session_id = self.session_id

        if self._should_encode_legacy_fields():
            if self.resource_id is not None:
                message.transfer_id = self.resource_id
            else:
                assert self.session_id != 0
                message.transfer_id = self.session_id

            # In the legacy protocol, the pending_bytes field must be set
            # alongside window_end_offset, as some transfer implementations
            # require it.
            if self.window_end_offset != 0:
                message.pending_bytes = self.window_end_offset - self.offset

        if self.data:
            message.data = self.data

        if self.remaining_bytes is not None:
            message.remaining_bytes = self.remaining_bytes

        if self.max_chunk_size_bytes is not None:
            message.max_chunk_size_bytes = self.max_chunk_size_bytes

        if self.min_delay_microseconds is not None:
            message.min_delay_microseconds = self.min_delay_microseconds

        if self.status is not None:
            message.status = self.status.value

        if self._is_initial_handshake_chunk():
            # During the initial handshake, the desired protocol version is
            # explictly encoded.
            message.protocol_version = self.protocol_version.value

        return message

    def id(self) -> int:
        """Returns the transfer context identifier for a chunk.

        Depending on the protocol version and type of chunk, this may correspond
        to one of several proto fields.
        """
        if self.resource_id is not None:
            # Always prioritize a resource_id over a session_id.
            return self.resource_id

        return self.session_id

    def requests_transmission_from_offset(self) -> bool:
        """Returns True if this chunk is requesting a retransmission."""
        return (
            self.type is Chunk.Type.PARAMETERS_RETRANSMIT
            or self.type is Chunk.Type.START
            or self.type is Chunk.Type.START_ACK_CONFIRMATION
        )

    def _is_initial_handshake_chunk(self) -> bool:
        return self.protocol_version is ProtocolVersion.VERSION_TWO and (
            self.type is Chunk.Type.START
            or self.type is Chunk.Type.START_ACK
            or self.type is Chunk.Type.START_ACK_CONFIRMATION
        )

    def _should_encode_legacy_fields(self) -> bool:
        return (
            self.protocol_version is ProtocolVersion.LEGACY
            or self.type is Chunk.Type.START
        )



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Inclusive language presubmit check."""

import dataclasses
from pathlib import Path
import re
from typing import Dict, List, Union

from . import presubmit

# List borrowed from Android:
# https://source.android.com/setup/contribute/respectful-code
# inclusive-language: disable
NON_INCLUSIVE_WORDS = [
    r'master',
    r'slave',
    r'red[-\s]?line',
    r'(white|gr[ae]y|black)[-\s]*(list|hat)',
    r'craz(y|ie)',
    r'insane',
    r'crip+led?',
    r'sanity',
    r'sane',
    r'dummy',
    r'grandfather',
    r's?he',
    r'his',
    r'her',
    r'm[ae]n[-\s]*in[-\s]*the[-\s]*middle',
    r'mitm',
    r'first[-\s]?class[-\s]?citizen',
]
# inclusive-language: enable

# Test: master  # inclusive-language: ignore
# Test: master


def _process_inclusive_language(*words):
    """Turn word list into one big regex with common inflections."""

    if not words:
        words = tuple(NON_INCLUSIVE_WORDS)

    all_words = []
    for entry in words:
        if isinstance(entry, str):
            all_words.append(entry)
        elif isinstance(entry, (list, tuple)):
            all_words.extend(entry)
        all_words.extend(x for x in words)
    all_words = tuple(all_words)

    # Confirm each individual word compiles as a valid regex.
    for word in all_words:
        _ = re.compile(word)

    word_boundary = (
        r'(\b|_|(?<=[a-z])(?=[A-Z])|(?<=[0-9])(?=\w)|(?<=\w)(?=[0-9]))'
    )

    return re.compile(
        r"({b})(?i:{w})(e?[sd]{b}|{b})".format(
            w='|'.join(all_words), b=word_boundary
        ),
    )


NON_INCLUSIVE_WORDS_REGEX = _process_inclusive_language()

# If seen, ignore this line and the next.
_IGNORE = 'inclusive-language: ignore'

# Ignore a whole section. Please do not change the order of these lines.
_DISABLE = 'inclusive-language: disable'
_ENABLE = 'inclusive-language: enable'


@dataclasses.dataclass
class PathMatch:
    word: str

    def __repr__(self):
        return f'Found non-inclusive word "{self.word}" in file path'


@dataclasses.dataclass
class LineMatch:
    line: int
    word: str

    def __repr__(self):
        return f'Found non-inclusive word "{self.word}" on line {self.line}'


@presubmit.check(name='inclusive_language')
def presubmit_check(
    ctx: presubmit.PresubmitContext,
    words_regex=NON_INCLUSIVE_WORDS_REGEX,
):
    """Presubmit check that ensures files do not contain banned words."""

    found_words: Dict[Path, List[Union[PathMatch, LineMatch]]] = {}

    for path in ctx.paths:
        match = words_regex.search(str(path.relative_to(ctx.root)))
        if match:
            found_words.setdefault(path, [])
            found_words[path].append(PathMatch(match.group(0)))

        if path.is_symlink() or path.is_dir():
            continue

        try:
            with open(path, 'r') as ins:
                enabled = True
                prev = ''
                for i, line in enumerate(ins, start=1):
                    if _DISABLE in line:
                        enabled = False
                    if _ENABLE in line:
                        enabled = True

                    # If we see the ignore line on this or the previous line we
                    # ignore any bad words on this line.
                    ignored = _IGNORE in prev or _IGNORE in line

                    if enabled and not ignored:
                        match = words_regex.search(line)

                        if match:
                            found_words.setdefault(path, [])
                            found_words[path].append(
                                LineMatch(i, match.group(0))
                            )

                    # Not using 'continue' so this line always executes.
                    prev = line

        except UnicodeDecodeError:
            # File is not text, like a gif.
            pass

    if found_words:
        with open(ctx.failure_summary_log, 'w') as outs:
            for i, (path, matches) in enumerate(found_words.items()):
                if i:
                    print('=' * 40, file=outs)
                print(path, file=outs)
                for match in matches:
                    print(match, file=outs)

        print(ctx.failure_summary_log.read_text(), end=None)

        print()
        print(
            """
Individual lines can be ignored with "inclusive-language: ignore". Blocks can be
ignored with "inclusive-language: disable" and reenabled with
"inclusive-language: enable".
""".strip()
        )
        # Re-enable just in case: inclusive-language: enable.

        raise presubmit.PresubmitFailure


def inclusive_language_checker(*words):
    """Create banned words checker for the given list of banned words."""

    regex = _process_inclusive_language(*words)

    def inclusive_language(  # pylint: disable=redefined-outer-name
        ctx: presubmit.PresubmitContext,
    ):
        globals()['inclusive_language'](ctx, regex)

    return inclusive_language



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""C++-related checks."""

import logging

from pw_presubmit import (
    build,
    Check,
    format_code,
    PresubmitContext,
    filter_paths,
)

_LOG: logging.Logger = logging.getLogger(__name__)


@filter_paths(endswith=format_code.CPP_HEADER_EXTS, exclude=(r'\.pb\.h$',))
def pragma_once(ctx: PresubmitContext) -> None:
    """Presubmit check that ensures all header files contain '#pragma once'."""

    for path in ctx.paths:
        _LOG.debug('Checking %s', path)
        with open(path) as file:
            for line in file:
                if line.startswith('#pragma once'):
                    break
            else:
                ctx.fail('#pragma once is missing!', path=path)


@Check
def asan(ctx: PresubmitContext) -> None:
    """Test with the address sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'asan')


@Check
def msan(ctx: PresubmitContext) -> None:
    """Test with the memory sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'msan')


@Check
def tsan(ctx: PresubmitContext) -> None:
    """Test with the thread sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'tsan')


@Check
def ubsan(ctx: PresubmitContext) -> None:
    """Test with the undefined behavior sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'ubsan')


@Check
def runtime_sanitizers(ctx: PresubmitContext) -> None:
    """Test with the address, thread, and undefined behavior sanitizers."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'runtime_sanitizers')


def all_sanitizers():
    # TODO(b/234876100): msan will not work until the C++ standard library
    # included in the sysroot has a variant built with msan.
    return [asan, tsan, ubsan, runtime_sanitizers]



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Text formatting functions."""

import copy
import re
from typing import Iterable, List, Tuple

from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple
from prompt_toolkit.utils import get_cwidth

_ANSI_SEQUENCE_REGEX = re.compile(r'\x1b[^m]*m')


def strip_ansi(text: str):
    """Strip out ANSI escape sequences."""
    return _ANSI_SEQUENCE_REGEX.sub('', text)


def split_lines(
    input_fragments: StyleAndTextTuples,
) -> List[StyleAndTextTuples]:
    """Break a flattened list of StyleAndTextTuples into a list of lines.

    Ending line breaks are not preserved."""
    lines: List[StyleAndTextTuples] = []
    this_line: StyleAndTextTuples = []
    for item in input_fragments:
        if item[1].endswith('\n'):
            # If there are no elements in this line except for a linebreak add
            # an empty StyleAndTextTuple so this line isn't an empty list.
            if len(this_line) == 0 and item[1] == '\n':
                this_line.append((item[0], item[1][:-1]))
            lines.append(this_line)
            this_line = []
        else:
            this_line.append(item)
    return lines


def insert_linebreaks(
    input_fragments: StyleAndTextTuples,
    max_line_width: int,
    truncate_long_lines: bool = True,
) -> Tuple[StyleAndTextTuples, int]:
    """Add line breaks at max_line_width if truncate_long_lines is True.

    Returns input_fragments with each character as it's own formatted text
    tuple."""
    fragments: StyleAndTextTuples = []
    total_width = 0
    line_width = 0
    line_height = 0
    new_break_inserted = False

    for item in input_fragments:
        # Check for non-printable fragment; doesn't affect the width.
        if '[ZeroWidthEscape]' in item[0]:
            fragments.append(item)
            continue

        new_item_style = item[0]

        # For each character in the fragment
        for character in item[1]:
            # Get the width respecting double width characters
            width = get_cwidth(character)
            # Increment counters
            total_width += width
            line_width += width
            # Save this character as it's own fragment
            if line_width <= max_line_width:
                if not new_break_inserted or character != '\n':
                    fragments.append((new_item_style, character))
                    # Was a line break just inserted?
                    if character == '\n':
                        # Increase height
                        line_height += 1
                new_break_inserted = False

            # Reset width to zero even if we are beyond the max line width.
            if character == '\n':
                line_width = 0

            # Are we at the limit for this line?
            elif line_width == max_line_width:
                # Insert a new linebreak fragment
                fragments.append((new_item_style, '\n'))
                # Increase height
                line_height += 1
                # Set a flag for skipping the next character if it is also a
                # line break.
                new_break_inserted = True

                if not truncate_long_lines:
                    # Reset line width to zero
                    line_width = 0

    # Check if the string ends in a final line break
    last_fragment_style = fragments[-1][0]
    last_fragment_text = fragments[-1][1]
    if not last_fragment_text.endswith('\n'):
        # Add a line break if none exists
        fragments.append((last_fragment_style, '\n'))
        line_height += 1

    return fragments, line_height


def join_adjacent_style_tuples(
    fragments: StyleAndTextTuples,
) -> StyleAndTextTuples:
    """Join adjacent FormattedTextTuples if they have the same style."""
    new_fragments: StyleAndTextTuples = []

    for i, fragment in enumerate(fragments):
        # Add the first fragment
        if i == 0:
            new_fragments.append(fragment)
            continue

        # Get this style
        style = fragment[0]
        # If the previous style matches
        if style == new_fragments[-1][0]:
            # Get the previous text
            new_text = new_fragments[-1][1]
            # Append this text
            new_text += fragment[1]
            # Replace the last fragment
            new_fragments[-1] = (style, new_text)
        else:
            # Styles don't match, just append.
            new_fragments.append(fragment)

    return new_fragments


def fill_character_width(
    input_fragments: StyleAndTextTuples,
    fragment_width: int,
    window_width: int,
    line_wrapping: bool = False,
    remaining_width: int = 0,
    horizontal_scroll_amount: int = 0,
    add_cursor: bool = False,
) -> StyleAndTextTuples:
    """Fill line to the width of the window using spaces."""
    # Calculate the number of spaces to add at the end.
    empty_characters = window_width - fragment_width
    # If wrapping is on, use remaining_width
    if line_wrapping and (fragment_width > window_width):
        empty_characters = remaining_width

    # Add additional spaces for horizontal scrolling.
    empty_characters += horizontal_scroll_amount

    if empty_characters <= 0:
        # No additional spaces required
        return input_fragments

    line_fragments = copy.copy(input_fragments)

    single_space = ('', ' ')
    line_ends_in_a_break = False
    # Replace the trailing \n with a space
    if line_fragments[-1][1] == '\n':
        line_fragments[-1] = single_space
        empty_characters -= 1
        line_ends_in_a_break = True

    # Append remaining spaces
    for _i in range(empty_characters):
        line_fragments.append(single_space)

    if line_ends_in_a_break:
        # Restore the \n
        line_fragments.append(('', '\n'))

    if add_cursor:
        # Add a cursor to this line by adding SetCursorPosition fragment.
        line_fragments_remainder = line_fragments
        line_fragments = [('[SetCursorPosition]', '')]
        # Use extend to keep types happy.
        line_fragments.extend(line_fragments_remainder)

    return line_fragments


def flatten_formatted_text_tuples(
    lines: Iterable[StyleAndTextTuples],
) -> StyleAndTextTuples:
    """Flatten a list of lines of FormattedTextTuples

    This function will also remove trailing newlines to avoid displaying extra
    empty lines in prompt_toolkit containers.
    """
    fragments: StyleAndTextTuples = []

    # Return empty list if lines is empty.
    if not lines:
        return fragments

    for line_fragments in lines:
        # Append all FormattedText tuples for this line.
        for fragment in line_fragments:
            fragments.append(fragment)

    # Strip off any trailing line breaks
    last_fragment: OneStyleAndTextTuple = fragments[-1]
    style = last_fragment[0]
    text = last_fragment[1].rstrip('\n')
    fragments[-1] = (style, text)
    return fragments


def remove_formatting(formatted_text: StyleAndTextTuples) -> str:
    """Throw away style info from prompt_toolkit formatted text tuples."""
    return ''.join([formatted_tuple[1] for formatted_tuple in formatted_text])


def get_line_height(text_width, screen_width, prefix_width):
    """Calculates line height for a string with line wrapping enabled."""
    if text_width == 0:
        return 0

    # If text will fit on the screen without wrapping.
    if text_width <= screen_width:
        return 1, screen_width - text_width

    # Assume zero width prefix if it's >= width of the screen.
    if prefix_width >= screen_width:
        prefix_width = 0

    # Start with height of 1 row.
    total_height = 1

    # One screen_width of characters (with no prefix) is displayed first.
    remaining_width = text_width - screen_width

    # While we have caracters remaining to be displayed
    while remaining_width > 0:
        # Add the new indentation prefix
        remaining_width += prefix_width
        # Display this line
        remaining_width -= screen_width
        # Add a line break
        total_height += 1

    # Remaining characters is what's left below zero.
    return (total_height, abs(remaining_width))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Help window container class."""

import functools
import importlib.resources
import inspect
import logging
from typing import Dict, Optional, TYPE_CHECKING

from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.layout import (
    ConditionalContainer,
    DynamicContainer,
    FormattedTextControl,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.widgets import Box, TextArea

from pygments.lexers.markup import RstLexer  # type: ignore
from pygments.lexers.data import YamlLexer  # type: ignore

from pw_console.style import (
    get_pane_indicator,
)
from pw_console.widgets import (
    mouse_handlers,
    to_keybind_indicator,
)

if TYPE_CHECKING:
    from pw_console.console_app import ConsoleApp

_LOG = logging.getLogger(__package__)

_PW_CONSOLE_MODULE = 'pw_console'


def _longest_line_length(text):
    """Return the longest line in the given text."""
    max_line_length = 0
    for line in text.splitlines():
        if len(line) > max_line_length:
            max_line_length = len(line)
    return max_line_length


class HelpWindow(ConditionalContainer):
    """Help window container for displaying keybindings."""

    # pylint: disable=too-many-instance-attributes

    def _create_help_text_area(self, **kwargs):
        help_text_area = TextArea(
            focusable=True,
            focus_on_click=True,
            scrollbar=True,
            style='class:help_window_content',
            wrap_lines=False,
            **kwargs,
        )

        # Additional keybindings for the text area.
        key_bindings = KeyBindings()
        register = self.application.prefs.register_keybinding

        @register('help-window.close', key_bindings)
        def _close_window(_event: KeyPressEvent) -> None:
            """Close the current dialog window."""
            self.toggle_display()

        if not self.disable_ctrl_c:

            @register('help-window.copy-all', key_bindings)
            def _copy_all(_event: KeyPressEvent) -> None:
                """Close the current dialog window."""
                self.copy_all_text()

        help_text_area.control.key_bindings = key_bindings
        return help_text_area

    def __init__(
        self,
        application: 'ConsoleApp',
        preamble: str = '',
        additional_help_text: str = '',
        title: str = '',
        disable_ctrl_c: bool = False,
    ) -> None:
        # Dict containing key = section title and value = list of key bindings.
        self.application: 'ConsoleApp' = application
        self.show_window: bool = False
        self.help_text_sections: Dict[str, Dict] = {}
        self._pane_title: str = title
        self.disable_ctrl_c = disable_ctrl_c

        # Tracks the last focused container, to enable restoring focus after
        # closing the dialog.
        self.last_focused_pane = None

        # Generated keybinding text
        self.preamble: str = preamble
        self.additional_help_text: str = additional_help_text
        self.help_text: str = ''

        self.max_additional_help_text_width: int = (
            _longest_line_length(self.additional_help_text)
            if additional_help_text
            else 0
        )
        self.max_description_width: int = 0
        self.max_key_list_width: int = 0
        self.max_line_length: int = 0

        self.help_text_area: TextArea = self._create_help_text_area()

        close_mouse_handler = functools.partial(
            mouse_handlers.on_click, self.toggle_display
        )
        copy_mouse_handler = functools.partial(
            mouse_handlers.on_click, self.copy_all_text
        )

        toolbar_padding = 1
        toolbar_title = ' ' * toolbar_padding
        toolbar_title += self.pane_title()

        buttons = []
        if not self.disable_ctrl_c:
            buttons.extend(
                to_keybind_indicator(
                    'Ctrl-c',
                    'Copy All',
                    copy_mouse_handler,
                    base_style='class:toolbar-button-active',
                )
            )
            buttons.append(('', '  '))

        buttons.extend(
            to_keybind_indicator(
                'q',
                'Close',
                close_mouse_handler,
                base_style='class:toolbar-button-active',
            )
        )
        top_toolbar = VSplit(
            [
                Window(
                    content=FormattedTextControl(
                        # [('', toolbar_title)]
                        functools.partial(
                            get_pane_indicator,
                            self,
                            toolbar_title,
                        )
                    ),
                    align=WindowAlign.LEFT,
                    dont_extend_width=True,
                ),
                Window(
                    content=FormattedTextControl([]),
                    align=WindowAlign.LEFT,
                    dont_extend_width=False,
                ),
                Window(
                    content=FormattedTextControl(buttons),
                    align=WindowAlign.RIGHT,
                    dont_extend_width=True,
                ),
            ],
            height=1,
            style='class:toolbar_active',
        )

        self.container = HSplit(
            [
                top_toolbar,
                Box(
                    body=DynamicContainer(lambda: self.help_text_area),
                    padding=Dimension(preferred=1, max=1),
                    padding_bottom=0,
                    padding_top=0,
                    char=' ',
                    style='class:frame.border',  # Same style used for Frame.
                ),
            ]
        )

        super().__init__(
            self.container,
            filter=Condition(lambda: self.show_window),
        )

    def pane_title(self):
        return self._pane_title

    def menu_title(self):
        """Return the title to display in the Window menu."""
        return self.pane_title()

    def __pt_container__(self):
        """Return the prompt_toolkit container for displaying this HelpWindow.

        This allows self to be used wherever prompt_toolkit expects a container
        object."""
        return self.container

    def copy_all_text(self):
        """Copy all text in the Python input to the system clipboard."""
        self.application.application.clipboard.set_text(
            self.help_text_area.buffer.text
        )

    def toggle_display(self):
        """Toggle visibility of this help window."""
        # Toggle state variable.
        self.show_window = not self.show_window

        if self.show_window:
            # Save previous focus
            self.last_focused_pane = self.application.focused_window()
            # Set the help window in focus.
            self.application.layout.focus(self.help_text_area)
        else:
            # Restore original focus if possible.
            if self.last_focused_pane:
                self.application.layout.focus(self.last_focused_pane)
            else:
                # Fallback to focusing on the first window pane.
                self.application.focus_main_menu()

    def content_width(self) -> int:
        """Return total width of help window."""
        # Widths of UI elements
        frame_width = 1
        padding_width = 1
        left_side_frame_and_padding_width = frame_width + padding_width
        right_side_frame_and_padding_width = frame_width + padding_width
        scrollbar_padding = 1
        scrollbar_width = 1

        desired_width = self.max_line_length + (
            left_side_frame_and_padding_width
            + right_side_frame_and_padding_width
            + scrollbar_padding
            + scrollbar_width
        )
        desired_width = max(60, desired_width)

        window_manager_width = (
            self.application.window_manager.current_window_manager_width
        )
        if not window_manager_width:
            window_manager_width = 80
        return min(desired_width, window_manager_width)

    def load_user_guide(self):
        rstdoc_text = importlib.resources.read_text(
            f'{_PW_CONSOLE_MODULE}.docs', 'user_guide.rst'
        )
        max_line_length = 0
        rst_text = ''
        for line in rstdoc_text.splitlines():
            if 'https://' not in line and len(line) > max_line_length:
                max_line_length = len(line)
            rst_text += line + '\n'
        self.max_line_length = max_line_length

        self.help_text_area = self._create_help_text_area(
            lexer=PygmentsLexer(RstLexer),
            text=rst_text,
        )

    def load_yaml_text(self, content: str):
        max_line_length = 0
        for line in content.splitlines():
            if 'https://' not in line and len(line) > max_line_length:
                max_line_length = len(line)
        self.max_line_length = max_line_length

        self.help_text_area = self._create_help_text_area(
            lexer=PygmentsLexer(YamlLexer),
            text=content,
        )

    def set_help_text(
        self, text: str, lexer: Optional[PygmentsLexer] = None
    ) -> None:
        self.help_text_area = self._create_help_text_area(
            lexer=lexer,
            text=text,
        )
        self._update_help_text_area(text)

    def generate_keybind_help_text(self) -> str:
        """Generate help text based on added key bindings."""

        template = self.application.get_template('keybind_list.jinja')

        text = template.render(
            sections=self.help_text_sections,
            max_additional_help_text_width=self.max_additional_help_text_width,
            max_description_width=self.max_description_width,
            max_key_list_width=self.max_key_list_width,
            preamble=self.preamble,
            additional_help_text=self.additional_help_text,
        )

        self._update_help_text_area(text)
        return text

    def _update_help_text_area(self, text: str) -> None:
        self.help_text = text

        # Find the longest line in the rendered template.
        self.max_line_length = _longest_line_length(self.help_text)

        # Replace the TextArea content.
        self.help_text_area.buffer.document = Document(
            text=self.help_text, cursor_position=0
        )

    def add_custom_keybinds_help_text(self, section_name, key_bindings: Dict):
        """Add hand written key_bindings."""
        self.help_text_sections[section_name] = key_bindings

    def add_keybind_help_text(self, section_name, key_bindings: KeyBindings):
        """Append formatted key binding text to this help window."""

        # Create a new keybind section, erasing any old section with thesame
        # title.
        self.help_text_sections[section_name] = {}

        # Loop through passed in prompt_toolkit key_bindings.
        for binding in key_bindings.bindings:
            # Skip this keybind if the method name ends in _hidden.
            if binding.handler.__name__.endswith('_hidden'):
                continue

            # Get the key binding description from the function doctstring.
            docstring = binding.handler.__doc__
            if not docstring:
                docstring = ''
            description = inspect.cleandoc(docstring)
            description = description.replace('\n', ' ')

            # Save the length of the description.
            if len(description) > self.max_description_width:
                self.max_description_width = len(description)

            # Get the existing list of keys for this function or make a new one.
            key_list = self.help_text_sections[section_name].get(
                description, list()
            )

            # Save the name of the key e.g. F1, q, ControlQ, ControlUp
            key_name = ' '.join(
                [getattr(key, 'name', str(key)) for key in binding.keys]
            )
            key_name = key_name.replace('Control', 'Ctrl-')
            key_name = key_name.replace('Shift', 'Shift-')
            key_name = key_name.replace('Escape ', 'Alt-')
            key_name = key_name.replace('Alt-Ctrl-', 'Ctrl-Alt-')
            key_name = key_name.replace('BackTab', 'Shift-Tab')
            key_list.append(key_name)

            key_list_width = len(', '.join(key_list))
            # Save the length of the key list.
            if key_list_width > self.max_key_list_width:
                self.max_key_list_width = key_list_width

            # Update this functions key_list
            self.help_text_sections[section_name][description] = key_list
